#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2005-2007  Donald N. Allingham
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2011       Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"ToolGeneration Framework"

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import logging

log = logging.getLogger(".")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.const import TOOL_OPTIONS
from gramps.gen.display.name import displayer as name_displayer
from gramps.gen.errors import WindowActiveError
from gramps.gen.plug._options import (
    Options,
    OptionHandler,
    OptionList,
    OptionListCollection,
)
from gramps.gen.plug import (
    TOOL_DEBUG,
    TOOL_ANAL,
    TOOL_DBPROC,
    TOOL_DBFIX,
    TOOL_REVCTL,
    TOOL_UTILS,
)

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

tool_categories = {
    TOOL_DEBUG: ("ToolDebug", _("Debug")),
    TOOL_ANAL: ("ToolAnExp", _("Analysis and Exploration")),
    TOOL_DBPROC: ("ToolProc", _("Family Tree Processing")),
    TOOL_DBFIX: ("ToolRep", _("Family Tree Repair")),
    TOOL_REVCTL: ("ToolRev", _("Revision Control")),
    TOOL_UTILS: ("ToolUtil", _("Utilities")),
}


# -------------------------------------------------------------------------
#
# Tool
#
# -------------------------------------------------------------------------
class Tool:
    """
    The Tool base class.  This is a base class for generating
    customized tools.  It cannot be used as is, but it can be easily
    sub-classed to create a functional tool.
    """

    def __init__(self, dbstate, options_class, name):
        from . import MenuToolOptions

        self.db = dbstate.db
        try:
            if issubclass(options_class, MenuToolOptions):
                # FIXME: pass in person_id
                self.options = options_class(name, None, dbstate)
            else:  # must be some kind of class or we get a TypeError
                self.options = options_class(name)
        except TypeError:
            self.options = options_class

        self.options.load_previous_values()
        if hasattr(options_class, "options_dict"):
            old_opts = options_class.saved_options_dict
            for key in options_class.options_dict:
                if options_class.options_dict[key] != old_opts[key]:
                    self.options.options_dict[key] = old_opts[key]

    def run_tool(self):
        pass


class BatchTool(Tool):
    """
    Same as Tool, except the warning is given about the potential
    loss of undo history.
    Should be used for tools using batch transactions.
    """

    def __init__(self, dbstate, user, options_class, name, parent=None):
        self._user = user
        if user.uistate:
            parent = user.uistate.window
        if not user.prompt(
            _("Undo history warning"),
            _(
                "Proceeding with this tool will erase the undo history "
                "for this session. In particular, you will not be able "
                "to revert the changes made by this tool or any changes "
                "made prior to it.\n\n"
                "If you think you may want to revert running this tool, "
                "please stop here and backup your database."
            ),
            _("_Proceed with the tool"),
            _("_Stop"),
            parent,
        ):
            self.fail = True
            return

        Tool.__init__(self, dbstate, options_class, name)
        self.fail = False


class ActivePersonTool(Tool):
    """
    Same as Tool , except the existence of the active person is checked
    and the tool is aborted if no active person exists. Should be used
    for tools that depend on active person.
    """

    def __init__(self, dbstate, uistate, options_class, name):
        if not uistate.get_active("Person"):
            # TODO: should we replace this with a callback?
            from ..dialog import ErrorDialog

            ErrorDialog(
                _("Active person has not been set"),
                _(
                    "You must select an active person for this "
                    "tool to work properly."
                ),
                parent=uistate.window,
            )
            self.fail = True
            return

        Tool.__init__(self, dbstate, options_class, name)
        self.fail = False


# ------------------------------------------------------------------------
#
# Command-line tool
#
# ------------------------------------------------------------------------
class CommandLineTool:
    """
    Provide a way to run tool from the command line.

    """

    def __init__(
        self, database, name, category, option_class, options_str_dict, noopt=False
    ):
        self.database = database
        self.category = category
        self.option_class = option_class(name)
        self.option_class.load_previous_values()
        self.show = options_str_dict.pop("show", None)
        self.options_str_dict = options_str_dict
        self.init_options(noopt)
        self.parse_option_str()
        self.show_options()

    def init_options(self, noopt):
        self.options_dict = {"id": ""}
        self.options_help = {
            "id": ["=ID", "Gramps ID of a central person."],
        }

        if noopt:
            return

        # Add tool-specific options
        for key in self.option_class.handler.options_dict:
            if key not in self.options_dict:
                self.options_dict[key] = self.option_class.handler.options_dict[key]

        # Add help for tool-specific options
        for key in self.option_class.options_help:
            if key not in self.options_help:
                self.options_help[key] = self.option_class.options_help[key]

    def parse_option_str(self):
        from gramps.cli.plug import _convert_str_to_match_type

        for opt in self.options_str_dict:
            if opt in self.options_dict:
                self.options_dict[opt] = _convert_str_to_match_type(
                    self.options_str_dict[opt], self.options_dict[opt]
                )
                self.option_class.handler.options_dict[opt] = self.options_dict[opt]
            else:
                print("Ignoring unknown option: %s" % opt)

        person_id = self.options_dict["id"]
        self.person = self.database.get_person_from_gramps_id(person_id)
        id_list = []

        for person in self.database.iter_people():
            id_list.append(
                "%s\t%s" % (person.get_gramps_id(), name_displayer.display(person))
            )
        self.options_help["id"].append(id_list)
        self.options_help["id"].append(False)

    def show_options(self):
        if not self.show:
            return
        elif self.show == "all":
            print("   Available options:")
            for key in self.options_dict:
                print("      %s" % key)
            print("   Use 'show=option' to see description and acceptable values")
        elif self.show in self.options_dict:
            print(
                "   %s%s\t%s"
                % (
                    self.show,
                    self.options_help[self.show][0],
                    self.options_help[self.show][1],
                )
            )
            print("   Available values are:")
            vals = self.options_help[self.show][2]
            if isinstance(vals, (list, tuple)):
                if self.options_help[self.show][3]:
                    for num, value in enumerate(vals):
                        print("      %d\t%s" % (num, value))
                else:
                    for val in vals:
                        print("      %s" % val)
            else:
                print("      %s" % self.options_help[self.show][2])

        else:
            self.show = None


# ------------------------------------------------------------------------
#
# Generic task functions for tools
#
# ------------------------------------------------------------------------
# Standard GUI tool generic task


def gui_tool(
    dbstate, user, tool_class, options_class, translated_name, name, category, callback
):
    """
    tool - task starts the report. The plugin system requires that the
    task be in the format of task that takes a database and a person as
    its arguments.
    """

    try:
        tool_class(
            dbstate=dbstate,
            user=user,
            options_class=options_class,
            name=name,
            callback=callback,
        )
    except WindowActiveError:
        pass
    except:
        log.error("Failed to start tool.", exc_info=True)


# Command-line generic task
def cli_tool(
    dbstate, name, category, tool_class, options_class, options_str_dict, user=None
):
    clt = CommandLineTool(dbstate.db, name, category, options_class, options_str_dict)
    clt.option_class.saved_options_dict = clt.options_dict

    # Exit here if show option was given
    if clt.show:
        return

    if user is None:
        from ...cli.user import User

        user = User()
    # run tool
    try:
        tool_class(
            dbstate=dbstate,
            user=user,
            options_class=clt.option_class,
            name=name,
            callback=None,
        )
    except:
        log.error("Failed to start tool.", exc_info=True)


# -------------------------------------------------------------------------
#
# Class handling options for plugins
#
# -------------------------------------------------------------------------
class ToolOptionHandler(OptionHandler):
    """
    Implements handling of the options for the plugins.
    """

    def __init__(self, module_name, options_dict, person_id=None):
        OptionHandler.__init__(self, module_name, options_dict, person_id)

    def init_subclass(self):
        self.collection_class = OptionListCollection
        self.list_class = OptionList
        self.filename = TOOL_OPTIONS


# ------------------------------------------------------------------------
#
# Tool Options class
#
# ------------------------------------------------------------------------
class ToolOptions(Options):
    """
    Defines options and provides handling interface.

    This is a base Options class for the tools. All tools, options
    classes should derive from it.
    """

    def __init__(self, name, person_id=None):
        """
        Initialize the class, performing usual house-keeping tasks.
        Subclasses MUST call this in their __init__() method.
        """
        self.name = name
        self.person_id = person_id
        self.options_dict = {}
        self.options_help = {}
        self.handler = None

    def load_previous_values(self):
        self.handler = ToolOptionHandler(self.name, self.options_dict, self.person_id)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import os
from io import StringIO
import html

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

_LOG = logging.getLogger(".DisplayState")

# -------------------------------------------------------------------------
#
# GNOME python modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import GObject
from gi.repository import GLib

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.utils.callback import Callback
from .utils import process_pending_events
from .views.navigationview import NavigationView
from gramps.gen.config import config
from gramps.gen.display.name import displayer as name_displayer
from .managedwindow import GrampsWindowManager
from gramps.gen.relationship import get_relationship_calculator
from .glade import Glade
from gramps.gen.utils.db import navigation_label
from gramps.gen.errors import HandleError
from .widgets.progressdialog import ProgressMonitor, GtkProgressDialog
from .dialog import ErrorDialog, WarningDialog
from .uimanager import ActionGroup
from .utils import get_display_size
from ..version import VERSION_QUALIFIER, DEV_VERSION
from gramps.gen.const import VERSION

DISABLED = -1


# -------------------------------------------------------------------------
#
# History manager
#
# -------------------------------------------------------------------------
class History(Callback):
    """History manages the objects of a certain type that have been viewed,
    with ability to go back, or forward.
    When accessing an object, it should be pushed on the History.
    """

    __signals__ = {"active-changed": (str,), "mru-changed": (list,)}

    def __init__(self, dbstate, nav_type):
        Callback.__init__(self)
        self.dbstate = dbstate
        self.nav_type = nav_type
        self.clear()

        dbstate.connect("database-changed", self.connect_signals)
        self.signal_map = {}
        self.signal_map[nav_type.lower() + "-delete"] = self.handles_removed
        self.signal_map[nav_type.lower() + "-rebuild"] = self.history_changed

    def connect_signals(self, dbstate):
        """
        Connects database signals when the database has changed.
        """
        for sig in self.signal_map:
            dbstate.connect(sig, self.signal_map[sig])

    def clear(self):
        """
        Clears the history, resetting the values back to their defaults
        """
        self.history = []
        self.mru = []
        self.index = -1
        self.lock = False

        if self.dbstate.is_open() and self.nav_type == "Person":
            initial_person = self.dbstate.db.find_initial_person()
            if initial_person:
                self.push(initial_person.get_handle())

    def push(self, handle):
        """
        Pushes the handle on the history stack
        """
        self.prune()
        if len(self.history) == 0 or handle != self.history[-1]:
            self.history.append(handle)
            if handle in self.mru:
                self.mru.remove(handle)
            self.mru.append(handle)
            self.emit("mru-changed", (self.mru,))
            self.index += 1
        if self.history:
            newact = self.history[self.index]
            if not isinstance(newact, str):
                newact = str(newact)
            self.emit("active-changed", (newact,))

    def forward(self, step=1):
        """
        Moves forward in the history list
        """
        self.index += step
        handle = self.history[self.index]
        if handle in self.mru:
            self.mru.remove(handle)
        self.mru.append(handle)
        self.emit("mru-changed", (self.mru,))
        newact = self.history[self.index]
        if not isinstance(newact, str):
            newact = str(newact)
        self.emit("active-changed", (newact,))
        return newact

    def back(self, step=1):
        """
        Moves backward in the history list
        """
        self.index -= step
        try:
            handle = self.history[self.index]
            if handle in self.mru:
                self.mru.remove(handle)
            self.mru.append(handle)
            self.emit("mru-changed", (self.mru,))
            newact = self.history[self.index]
            if not isinstance(newact, str):
                newact = str(newact)
            self.emit("active-changed", (newact,))
            return newact
        except IndexError:
            return ""

    def present(self):
        """
        return the person handle that is now active in the history
        """
        try:
            if self.history:
                return self.history[self.index]
            else:
                return ""
        except IndexError:
            return ""

    def at_end(self):
        """
        returns True if we are at the end of the history list
        """
        return self.index + 1 == len(self.history)

    def at_front(self):
        """
        returns True if we are at the front of the history list
        """
        return self.index <= 0

    def prune(self):
        """
        Truncates the history list at the current object.
        """
        if not self.at_end():
            self.history = self.history[0 : self.index + 1]

    def handles_removed(self, handle_list):
        """
        Called in response to an object-delete signal.
        Removes a list of handles from the history.
        """
        for del_id in handle_list:
            history_count = self.history.count(del_id)
            for dummy in range(history_count):
                self.history.remove(del_id)
                self.index -= 1

            mhc = self.mru.count(del_id)
            for dummy in range(mhc):
                self.mru.remove(del_id)
        if self.history:
            newact = self.history[self.index]
            if not isinstance(newact, str):
                newact = str(newact)
            self.emit("active-changed", (newact,))
        self.emit("mru-changed", (self.mru,))

    def history_changed(self):
        """
        Called in response to an object-rebuild signal.
        Objects in the history list may have been deleted.
        """
        self.clear()
        self.emit("mru-changed", (self.mru,))


# -------------------------------------------------------------------------
#
# Recent Docs Menu
#
# -------------------------------------------------------------------------

_RCT_TOP = '<placeholder id="OpenRecentMenu">'
_RCT_MENU = """
            <item>
              <attribute name="action">win.%s</attribute>
              <attribute name="label">%s</attribute>
            </item>"""
_RCT_BTM = "\n          </placeholder>\n"
_RCT_BAR_TOP = (
    '<object class="GtkMenu"  id="OpenBtnMenu">\n'
    '<property name="visible">True</property>\n'
    '<property name="can_focus">False</property>'
)
_RCT_BAR = """
<child>
    <object class="GtkMenuItem">
        <property name="action-name">win.%s</property>
        <property name="label">%s</property>
        <property name="use_underline">False</property>
        <property name="visible">True</property>
    </object>
</child>"""
_RCT_BAR_BTM = "\n</object>\n"


from gramps.gen.recentfiles import RecentFiles


class RecentDocsMenu:
    def __init__(self, uistate, state, fileopen):
        self.ui_xml = []
        self.action_group = ActionGroup("RecentFiles")
        self.active = DISABLED
        self.uistate = uistate
        self.uimanager = uistate.uimanager
        self.fileopen = fileopen
        self.state = state

    def load(self, item):
        filename = item.get_path()
        try:
            self.fileopen(filename)
        except Exception as err:
            ErrorDialog(_("Cannot load database"), str(err), parent=self.uistate.window)

    def build(self, update_menu=True):
        gramps_rf = RecentFiles()

        count = 0

        if self.active != DISABLED:
            self.uimanager.remove_ui(self.ui_xml)
            self.uimanager.remove_action_group(self.action_group)
            self.active = DISABLED

        actionlist = []
        menu = _RCT_TOP
        bar = _RCT_BAR_TOP
        rfiles = gramps_rf.gramps_recent_files
        rfiles.sort(key=lambda x: x.get_time(), reverse=True)

        # new_menu = Gtk.Menu()
        # new_menu.set_tooltip_text(_("Connect to a recent database"))

        for item in rfiles:
            try:
                title = html.escape(item.get_name())
                filename = os.path.basename(item.get_path())
                action_id = "RecentMenu%d" % count
                # add the menuitem for this file
                menu += _RCT_MENU % (action_id, title)
                # add the action for this file
                actionlist.append((action_id, make_callback(item, self.load)))
                # add the toolbar menuitem
                bar += _RCT_BAR % (action_id, title)
            except RuntimeError:
                # ignore no longer existing files
                _LOG.info("Ignoring the RecentItem %s (%s)" % (title, filename))

            count += 1
        menu += _RCT_BTM
        bar += _RCT_BAR_BTM
        self.ui_xml = [menu, bar]
        self.action_group.add_actions(actionlist)
        self.uimanager.insert_action_group(self.action_group)
        self.active = self.uimanager.add_ui_from_string(self.ui_xml)
        if update_menu:
            self.uimanager.update_menu()


def make_callback(val, func):
    return lambda x, y: func(val)


from .logger import RotateHandler


class WarnHandler(RotateHandler):
    def __init__(self, capacity, button, parent=None):
        RotateHandler.__init__(self, capacity)
        self.setLevel(logging.WARN)
        self.button = button
        button.on_clicked(self.display)
        self.timer = None
        self.last_line = "-1"
        self.parent = parent

    def emit(self, record):
        if self.timer is None:
            # check every 3 minutes if warn button can disappear
            self.timer = GLib.timeout_add(3 * 60 * 1000, self._check_clear)
        RotateHandler.emit(self, record)
        self.button.show()

    def _check_clear(self):
        buffer = self.get_buffer()
        if len(buffer) > 0:
            new_last_line = self.get_buffer()[-1]
            if self.last_line == new_last_line:
                # buffer has not changed for 3 minutes, let's clear it:
                self._clear()
                return False
            else:
                self.last_line = new_last_line
                return True
        else:
            return False

    def _clear(self):
        self.button.hide()
        self.set_capacity(self._capacity)
        self.last_line = "-1"
        self.timer = None

    def display(self, obj):
        obj.hide()
        self.glade = Glade(toplevel="displaystate")
        top = self.glade.toplevel
        msg = self.glade.get_object("msg")
        buf = msg.get_buffer()
        for i in self.get_formatted_log():
            buf.insert_at_cursor(i + "\n")
        if self.parent:
            top.set_transient_for(self.parent)
        top.run()
        top.destroy()


TOOL_UI = """  <child>
    <object class="GtkToolButton">
      <property name="icon-name">%s</property>
      <property name="action-name">%s</property>
      <property name="tooltip_text">%s</property>
      <property name="label">%s</property>
      <property name="use-underline">True</property>
    </object>
    <packing>
      <property name="homogeneous">False</property>
    </packing>
  </child>
"""

TOOLS = {
    "clipboard": (
        "edit-paste",
        "win.Clipboard",
        _("Open the Clipboard dialog"),
        _("Clip_board"),
    ),
    "reports": (
        "gramps-reports",
        "win.Reports",
        _("Open the reports dialog"),
        _("Reports"),
    ),
    "tools": ("gramps-tools", "win.Tools", _("Open the tools dialog"), _("Tools")),
    "addons": (
        "gramps-addon",
        "win.AddonManager",
        _("Open Addon Manager"),
        _("Addons"),
    ),
    "preference": (
        "gramps-preferences",
        "app.preferences",
        _("Open Preferences"),
        _("Preferences"),
    ),
}


class DisplayState(Callback):
    __signals__ = {
        "filters-changed": (str,),
        "filter-name-changed": (str, str, str),
        "nameformat-changed": None,
        "placeformat-changed": None,
        "grampletbar-close-changed": None,
        "update-available": (list,),
        "autobackup": None,
        "font-changed": None,
        "toolbar-changed": None,
    }

    # nav_type to message
    NAV2MES = {
        "Person": _("No active person"),
        "Family": _("No active family"),
        "Event": _("No active event"),
        "Place": _("No active place"),
        "Source": _("No active source"),
        "Citation": _("No active citation"),
        "Repository": _("No active repository"),
        "Media": _("No active media"),
        "Note": _("No active note"),
    }

    BUSY_CURSOR = Gdk.Cursor.new_for_display(
        Gdk.Display.get_default(), Gdk.CursorType.WATCH
    )

    def __init__(self, window, status, uimanager, viewmanager=None):
        self.busy = False
        self.cursor = None
        self.viewmanager = viewmanager
        self.uimanager = uimanager
        self.progress_monitor = ProgressMonitor(GtkProgressDialog, ("", window))
        self.window = window
        Callback.__init__(self)
        self.status = status
        self.status_id = status.get_context_id("GRAMPS")
        self.progress = status.get_progress_bar()
        self.status_ver = status.get_version_btn()
        self.history_lookup = {}
        self.gwm = GrampsWindowManager(uimanager)
        self.widget = None
        self.disprel_old = ""
        self.disprel_defpers = None
        self.disprel_active = None
        self.set_relationship_class()
        self.export = False
        self.backup_timer = None
        self.symbols = config.get("utf8.in-use")
        self.death_symbol = config.get("utf8.death-symbol")

        formatter = logging.Formatter("%(levelname)s %(name)s: %(message)s")
        warnbtn = status.get_warning_button()
        self.rhandler = WarnHandler(capacity=400, button=warnbtn, parent=window)
        self.rhandler.setFormatter(formatter)
        self.rhandler.setLevel(logging.WARNING)
        self.log = logging.getLogger()
        self.log.addHandler(self.rhandler)
        # This call has been moved one level up,
        # but this connection is still made!
        # self.dbstate.connect('database-changed', self.db_changed)
        self.connect("toolbar-changed", self.set_toolbar)

        if DEV_VERSION or VERSION_QUALIFIER:
            ver_btn = status.get_version_btn()
            ver_btn.set_label(VERSION)
            if DEV_VERSION:
                msg = "master"
            else:
                msg = VERSION_QUALIFIER[1:]
            ver_btn.connect("clicked", self.__develop_warn, msg)
            ver_btn.show()

    def set_toolbar(self):
        ui = '<placeholder id="AfterTools">\n'
        for key in TOOLS.keys():
            if config.get("interface.toolbar-" + key):
                ui += TOOL_UI % TOOLS[key]
        ui += "</placeholder>"
        self.uimanager.add_ui_from_string([ui])
        self.uimanager.update_menu()

    def set_backup_timer(self):
        """
        Set the backup timer.
        """
        interval = config.get("database.autobackup")
        if self.backup_timer is not None:
            GLib.source_remove(self.backup_timer)
            self.backup_timer = None
        if interval == 1:
            minutes = 15
        elif interval == 2:
            minutes = 30
        elif interval == 3:
            minutes = 60
        elif interval == 4:
            minutes = 720
        elif interval == 5:
            minutes = 1440
        if interval > 0:
            self.backup_timer = GLib.timeout_add_seconds(
                minutes * 60, self.__emit_autobackup
            )

    def __emit_autobackup(self):
        """
        Emit an 'autobackup' signal.
        """
        self.emit("autobackup")
        return True

    def screen_width(self):
        """
        Return the width of the current screen.
        """
        width, _height = get_display_size(self.window)
        return width

    def screen_height(self):
        """
        Return the height of the current screen.
        """
        _width, height = get_display_size(self.window)
        return height

    def clear_history(self):
        """
        Clear all history objects.
        """
        for history in list(self.history_lookup.values()):
            history.clear()

    def get_history(self, nav_type, nav_group=0):
        """
        Return the history object for the given navigation type and group.
        """
        return self.history_lookup.get((nav_type, nav_group))

    def register(self, dbstate, nav_type, nav_group):
        """
        Create a history and navigation object for the specified
        navigation type and group, if they don't exist.
        """
        if (nav_type, nav_group) not in self.history_lookup:
            history = History(dbstate, nav_type)
            self.history_lookup[(nav_type, nav_group)] = history

    def get_active(self, nav_type, nav_group=0):
        """
        Return the handle for the active obejct specified by the given
        navigation type and group.
        """
        history = self.get_history(nav_type, nav_group)
        return history.present() if history else None

    def set_active(self, handle, nav_type, nav_group=0):
        """
        Set the active object for the specified navigation type and group to
        the given handle.
        """
        history = self.get_history(nav_type, nav_group)
        if history:
            history.push(handle)

    def set_sensitive(self, state):
        tbar = self.uimanager.get_widget("ToolBar")
        tbar.set_sensitive(state)
        self.viewmanager.hpane.set_sensitive(state)
        self.uimanager.enable_all_actions(state)

    def db_changed(self, db):
        db.connect("long-op-start", self.progress_monitor.add_op)
        self.clear_history()

    def set_relationship_class(self):
        """method that rebinds the relationship to the current rel calc
        Should be called after load or reload of plugins
        """
        self.relationship = get_relationship_calculator(reinit=True)

    def set_gendepth(self, value):
        """Set the generations we search back for showing relationships
        on Gramps interface. Value must be integer > 0
        This method will be used by the preference editor when user changes
        the generations.
        """
        self.relationship.set_depth(value)

    def display_relationship(self, dbstate, active_handle):
        """Construct the relationship in order to show it in the statusbar
        This can be a time intensive calculation, so we only want to do
        it if persons are different than before.
        Eg: select a person, then double click, will result in calling
            three times to construct build the statusbar. We only want
            to obtain relationship once!
        This means the relationship part of statusbar only changes on
        change of row.
        """
        self.relationship.connect_db_signals(dbstate)
        default_person = dbstate.db.get_default_person()
        if default_person is None or active_handle is None:
            return ""
        if (
            default_person.handle == self.disprel_defpers
            and active_handle == self.disprel_active
        ):
            return self.disprel_old

        active = dbstate.db.get_person_from_handle(active_handle)
        if active is None:
            # During merger this method can be called at a time when treemodel
            # and database are not in sync, resulting in active_handle != None,
            # but active == None; see bug 5290 for the details.
            return ""
        name = self.relationship.get_one_relationship(
            dbstate.db, default_person, active
        )
        # store present call data
        self.disprel_old = name
        self.disprel_defpers = default_person.handle
        self.disprel_active = active_handle
        if name:
            return name
        else:
            return ""

    def set_export_mode(self, value):
        self.set_busy_cursor(value)
        if value == self.export:
            return
        else:
            self.export = value

    def get_export_mode(self):
        return self.export

    def set_busy_cursor(self, value):
        if value == self.busy:
            return
        else:
            self.busy = value
        if self.window.get_window():
            if value:
                self.cursor = self.window.get_window().get_cursor()
                self.window.get_window().set_cursor(self.BUSY_CURSOR)
            else:
                self.window.get_window().set_cursor(self.cursor)
                if self.window.get_window().is_visible():
                    # avoid critical gdk error:
                    # Gdk-CRITICAL **: gdk_error_trap_pop_internal: assertion `trap != NULL' failed
                    # only process events if window is actually visible
                    process_pending_events()

    def set_open_widget(self, widget):
        self.widget = widget

    def set_open_recent_menu(self, menu):
        self.widget.set_menu(menu)

    def push_message(self, dbstate, text):
        self.status_text(text)
        GLib.timeout_add(5000, self.modify_statusbar, dbstate)

    def show_filter_results(self, dbstate, matched, total):
        # nav_type = self.viewmanager.active_page.navigation_type()
        # text = ((_("%(nav_type)s View") % {"nav_type": _(nav_type)}) +
        text = self.viewmanager.active_page.get_title() + (": %d/%d" % (matched, total))
        self.status.set_filter(text)

    def clear_filter_results(self):
        self.status.clear_filter()

    def modify_statusbar(self, dbstate, active=None):
        """Update the status bar with current object info.

        Since this is called via GLib.timeout_add it can happen at any time
        Gtk is idle or processing pending events.  Even in the midst of a
        multiple delete, before the GUI has been updated for missing objects.
        So it is susceptible to HandleErrors for missing data, thus the 'try'.
        """
        try:
            view = self.viewmanager.active_page
            if not isinstance(view, NavigationView) or dbstate is None:
                return

            nav_type = view.navigation_type()
            active_handle = self.get_active(nav_type, view.navigation_group())

            self.status.pop(self.status_id)

            if active_handle and dbstate.is_open():
                name, _obj = navigation_label(dbstate.db, nav_type, active_handle)
                # Append relationship to default person if enabled.
                if nav_type == "Person" and config.get("interface.statusbar") > 1:
                    if active_handle != dbstate.db.get_default_handle():
                        msg = self.display_relationship(dbstate, active_handle)
                        if msg:
                            name = "%s (%s)" % (name, msg.strip())
            else:
                name = _("No active object")

            if not name:
                name = self.NAV2MES[nav_type]

            self.status.push(self.status_id, name)
            process_pending_events()
        except HandleError:
            return

    def pulse_progressbar(self, value, text=None):
        self.progress.set_fraction(min(value / 100.0, 1.0))
        if text:
            self.progress.set_text("%s: %d%%" % (text, value))
        else:
            self.progress.set_text("%d%%" % value)
        process_pending_events()

    def status_text(self, text):
        self.status.pop(self.status_id)
        self.status.push(self.status_id, text)
        process_pending_events()

    def reload_symbols(self):
        self.symbols = config.get("utf8.in-use")
        self.death_symbol = config.get("utf8.death-symbol")

    def __develop_warn(self, button, warning_type):
        """
        Display a development warning message to the user, with the
        warning_type in it.

        :param warning_type: the general name of the warning, e.g. "master"
        :type warning_type: str
        """
        WarningDialog(
            _("Danger: This is unstable code!"),
            _("This Gramps ('%s') is a development release.\n") % warning_type
            + _(
                "This version is not meant for normal usage. Use "
                "at your own risk.\n"
                "\n"
                "This version may:\n"
                "1) Work differently than you expect.\n"
                "2) Fail to run at all.\n"
                "3) Crash often.\n"
                "4) Corrupt your data.\n"
                "5) Save data in a format that is incompatible with the "
                "official release.\n"
                "\n"
                "%(bold_start)sBACKUP%(bold_end)s "
                "your existing databases before opening "
                "them with this version, and make sure to export your "
                "data to XML every now and then."
            )
            % {"bold_start": "<b>", "bold_end": "</b>"},
            parent=self.window,
        )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""File and File format management for the different reports
"""

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".docbackend.py")

# ------------------------------------------------------------------------
#
# Functions
#
# ------------------------------------------------------------------------


def noescape(text):
    """
    Function that does not escape the text passed. Default for backends
    """
    return text


# -------------------------------------------------------------------------
#
# DocBackend exception
#
# -------------------------------------------------------------------------
class DocBackendError(Exception):
    """Error used to report docbackend errors."""

    def __init__(self, value=""):
        Exception.__init__(self)
        self.value = value

    def __str__(self):
        return self.value


# ------------------------------------------------------------------------
#
# Document Backend class
#
# ------------------------------------------------------------------------


class DocBackend:
    """
    Base class for text document backends.
    The DocBackend manages a file to which it writes. It further knowns
    enough of the file format to be able to translate between the BaseDoc API
    and the file format.
    Specifically for text reports a translation of styled notes to the file
    format usage is done.
    """

    BOLD = 0
    ITALIC = 1
    UNDERLINE = 2
    FONTFACE = 3
    FONTSIZE = 4
    FONTCOLOR = 5
    HIGHLIGHT = 6
    SUPERSCRIPT = 7
    LINK = 8
    STRIKETHROUGH = 9
    SUBSCRIPT = 10

    SUPPORTED_MARKUP = []

    ESCAPE_FUNC = lambda: noescape
    # Map between styletypes and internally used values. This map is needed
    # to make TextDoc officially independant of gen.lib.styledtexttag
    STYLETYPE_MAP = {}
    CLASSMAP = None

    # STYLETAGTABLE to store markup for write_markup associated with style tags
    STYLETAG_MARKUP = {
        BOLD: ("", ""),
        ITALIC: ("", ""),
        UNDERLINE: ("", ""),
        SUPERSCRIPT: ("", ""),
        LINK: ("", ""),
        STRIKETHROUGH: ("", ""),
        SUBSCRIPT: ("", ""),
    }

    def __init__(self, filename=None):
        """
        :param filename: path name of the file the backend works on
        """
        self.__file = None
        self._filename = filename

    def getf(self):
        """
        Obtain the filename on which backend writes
        """
        return self._filename

    def setf(self, value):
        """
        Set the filename on which the backend writes, changing the value
        passed on initialization.
        Can only be done if the previous filename is not open.
        """
        if self.__file is not None:
            raise ValueError(_("Close file first"))
        self._filename = value

    filename = property(getf, setf, None, "The filename the backend works on")

    def open(self):
        """
        Opens the document.
        """
        if self.filename is None:
            raise DocBackendError(_("No filename given"))
        if self.__file is not None:
            raise DocBackendError(
                _("File %s already open, close it first.") % self.filename
            )
        self._checkfilename()
        try:
            self.__file = open(self.filename, "w", encoding="utf-8")
        except IOError as msg:
            errmsg = "%s\n%s" % (_("Could not create %s") % self.filename, msg)
            raise DocBackendError(errmsg)
        except:
            raise DocBackendError(_("Could not create %s") % self.filename)

    def _checkfilename(self):
        """
        Check to make sure filename satisfies the standards for this filetype
        """
        pass

    def close(self):
        """
        Closes the file that is written on.
        """
        if self.__file is None:
            raise IOError("No file open")
        self.__file.close()
        self.__file = None

    def write(self, string):
        """
        Write a string to the file. There is no return value.
        Due to buffering, the string may not actually show up until the
        :meth:`close` method is called.
        """
        self.__file.write(string)

    def writelines(self, sequence):
        """
        Write a sequence of strings to the file. The sequence can be any
        iterable object producing strings, typically a list of strings.
        """
        self.__file.writelines(sequence)

    def escape(self, preformatted=False):
        """
        The escape func on text for this file format.

        :param preformatted: some formats can have different escape function
                             for normal text and preformatted text
        :type preformatted: bool
        """
        return self.ESCAPE_FUNC()

    def find_tag_by_stag(self, s_tag):
        """
        :param s_tag: object: assumed styledtexttag

        A styled tag is type with a value. Every styled tag must be converted
        to the tags used in the corresponding markup for the backend,
        eg <b>text</b> for bold in html. These markups are stored in
        STYLETAG_MARKUP. They are tuples for begin and end tag. If a markup is
        not present yet, it is created, using the :meth:`_create_xmltag` method
        you can overwrite.
        """
        tagtype = s_tag.name

        if not self.STYLETYPE_MAP or self.CLASSMAP != tagtype.__class__.__name__:
            self.CLASSMAP = tagtype.__class__.__name__
            self.STYLETYPE_MAP[tagtype.BOLD] = self.BOLD
            self.STYLETYPE_MAP[tagtype.ITALIC] = self.ITALIC
            self.STYLETYPE_MAP[tagtype.UNDERLINE] = self.UNDERLINE
            self.STYLETYPE_MAP[tagtype.FONTFACE] = self.FONTFACE
            self.STYLETYPE_MAP[tagtype.FONTSIZE] = self.FONTSIZE
            self.STYLETYPE_MAP[tagtype.FONTCOLOR] = self.FONTCOLOR
            self.STYLETYPE_MAP[tagtype.HIGHLIGHT] = self.HIGHLIGHT
            self.STYLETYPE_MAP[tagtype.SUPERSCRIPT] = self.SUPERSCRIPT
            self.STYLETYPE_MAP[tagtype.LINK] = self.LINK
            self.STYLETYPE_MAP[tagtype.STRIKETHROUGH] = self.STRIKETHROUGH
            self.STYLETYPE_MAP[tagtype.SUBSCRIPT] = self.SUBSCRIPT

        if s_tag.name == tagtype.LINK:
            return self.format_link(s_tag.value)
        typeval = int(s_tag.name)
        s_tagvalue = s_tag.value
        tag_name = None
        if tagtype.STYLE_TYPE[typeval] == bool:
            return self.STYLETAG_MARKUP[self.STYLETYPE_MAP[typeval]]
        elif tagtype.STYLE_TYPE[typeval] == str:
            tag_name = "%d %s" % (typeval, s_tagvalue)
        elif tagtype.STYLE_TYPE[typeval] == int:
            tag_name = "%d %d" % (typeval, int(s_tagvalue))
        if not tag_name:
            return None

        tags = self.STYLETAG_MARKUP.get(tag_name)
        if tags is not None:
            return tags
        # no tag known yet, create the markup, add to lookup, and return
        tags = self._create_xmltag(self.STYLETYPE_MAP[typeval], s_tagvalue)
        self.STYLETAG_MARKUP[tag_name] = tags
        return tags

    def _create_xmltag(self, tagtype, value):
        """
        Create the xmltags for the backend.
        Overwrite this method to create functionality with a backend
        """
        if tagtype not in self.SUPPORTED_MARKUP:
            return None
        return ("", "")

    def add_markup_from_styled(self, text, s_tags, split="", escape=True):
        """
        Input is plain text, output is text with markup added according to the
        s_tags which are assumed to be styledtexttags.
        When split is given the text will be split over the value given, and
        tags applied in such a way that it the text can be safely splitted in
        pieces along split.

        :param text: str, a piece of text
        :param s_tags: styledtexttags that must be applied to the text
        :param split: str, optional. A string along which the output can
                      be safely split without breaking the styling.

        As adding markup means original text must be escaped, ESCAPE_FUNC is
        used. This can be used to convert the text of a styledtext to the format
        needed for a document backend.  Do not call this method in a report,
        use the :meth:`write_markup` method.

        .. note:: the algorithm is complex as it assumes mixing of tags is not
                  allowed: eg <b>text<i> here</b> not</i> is assumed invalid
                  as markup. If the s_tags require such a setup, what is
                  returned is <b>text</b><i><b> here</b> not</i>
                  overwrite this method if this complexity is not needed.
        """
        if not escape:
            escape_func = self.ESCAPE_FUNC
            self.ESCAPE_FUNC = lambda: (lambda text: text)
        # unicode text must be sliced correctly
        text = str(text)
        FIRST = 0
        LAST = 1
        tagspos = {}
        for s_tag in s_tags:
            tag = self.find_tag_by_stag(s_tag)
            if tag is not None:
                for start, end in s_tag.ranges:
                    if start in tagspos:
                        tagspos[start] += [(tag, FIRST)]
                    else:
                        tagspos[start] = [(tag, FIRST)]
                    if end in tagspos:
                        tagspos[end] = [(tag, LAST)] + tagspos[end]
                    else:
                        tagspos[end] = [(tag, LAST)]
        start = 0
        end = len(text)
        keylist = list(tagspos.keys())
        keylist.sort()
        keylist = [x for x in keylist if x <= len(text)]
        opentags = []
        otext = ""  # the output, text with markup
        lensplit = len(split)
        for pos in keylist:
            # write text up to tag
            if pos > start:
                if split:
                    # make sure text can split
                    splitpos = text[start:pos].find(split)
                    while splitpos != -1:
                        otext += self.ESCAPE_FUNC()(text[start : start + splitpos])
                        # close open tags
                        for opentag in reversed(opentags):
                            otext += opentag[1]
                        # add split text
                        otext += self.ESCAPE_FUNC()(split)
                        # open the tags again
                        for opentag in opentags:
                            otext += opentag[0]
                        # obtain new values
                        start = start + splitpos + lensplit
                        splitpos = text[start:pos].find(split)

                otext += self.ESCAPE_FUNC()(text[start:pos])
            # write out tags
            for tag in tagspos[pos]:
                # close open tags starting from last open
                for opentag in reversed(opentags):
                    otext += opentag[1]
                # if start, add to opentag in beginning as first to open
                if tag[1] == FIRST:
                    opentags = [tag[0]] + opentags
                else:
                    # end tag, is closed already, remove from opentag
                    opentags = [x for x in opentags if not x == tag[0]]
                # now all tags are closed, open the ones that should open
                for opentag in opentags:
                    otext += opentag[0]
            start = pos
        # add remainder of text, no markup present there if all is correct
        if opentags:
            # a problem, we don't have a closing tag left but there are open
            # tags. Just keep them up to end of text
            pos = len(text)
            print(
                "WARNING: DocBackend : More style tags in text than length "
                "of text allows.\n",
                opentags,
            )
            if pos > start:
                if split:
                    # make sure text can split
                    splitpos = text[start:pos].find(split)
                    while splitpos != -1:
                        otext += self.ESCAPE_FUNC()(text[start : start + splitpos])
                        # close open tags
                        for opentag in reversed(opentags):
                            otext += opentag[1]
                        # add split text
                        otext += self.ESCAPE_FUNC()(split)
                        # open the tags again
                        for opentag in opentags:
                            otext += opentag[0]
                        # obtain new values
                        start = start + splitpos + lensplit
                        splitpos = text[start:pos].find(split)

                otext += self.ESCAPE_FUNC()(text[start:pos])
            for opentag in reversed(opentags):
                otext += opentag[1]
        else:
            otext += self.ESCAPE_FUNC()(text[start:end])
        if not escape:
            self.ESCAPE_FUNC = escape_func
        return otext

    def format_link(self, value):
        """
        Default format for links. Override for better support.

        value is: "TYPE DATA" where TYPE is 'url' or 'ref'.

        """
        return self.STYLETAG_MARKUP[DocBackend.UNDERLINE]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2010       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
LocationBase class for Gramps.
"""


# -------------------------------------------------------------------------
#
# LocationBase class
#
# -------------------------------------------------------------------------
class LocationBase:
    """
    Base class for all things Address.
    """

    def __init__(self, source=None):
        """
        Create a LocationBase object, copying from the source object if it
        exists.
        """
        if source:
            self.street = source.street
            self.locality = source.locality
            self.city = source.city
            self.county = source.county
            self.state = source.state
            self.country = source.country
            self.postal = source.postal
            self.phone = source.phone
        else:
            self.street = ""
            self.locality = ""
            self.city = ""
            self.county = ""
            self.state = ""
            self.country = ""
            self.postal = ""
            self.phone = ""

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return (
            self.street,
            self.locality,
            self.city,
            self.county,
            self.state,
            self.country,
            self.postal,
            self.phone,
        )

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        (
            self.street,
            self.locality,
            self.city,
            self.county,
            self.state,
            self.country,
            self.postal,
            self.phone,
        ) = data
        return self

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [
            self.street,
            self.locality,
            self.city,
            self.county,
            self.state,
            self.country,
            self.postal,
            self.phone,
        ]

    def set_street(self, val):
        """Set the street portion of the Location."""
        self.street = val

    def get_street(self):
        """Return the street portion of the Location."""
        return self.street

    def set_locality(self, val):
        """Set the locality portion of the Location."""
        self.locality = val

    def get_locality(self):
        """Return the locality portion of the Location."""
        return self.locality

    def set_city(self, data):
        """Set the city name of the LocationBase object."""
        self.city = data

    def get_city(self):
        """Return the city name of the LocationBase object."""
        return self.city

    def set_postal_code(self, data):
        """Set the postal code of the LocationBase object."""
        self.postal = data

    def get_postal_code(self):
        """Return the postal code of the LocationBase object."""
        return self.postal

    def set_phone(self, data):
        """Set the phone number of the LocationBase object."""
        self.phone = data

    def get_phone(self):
        """Return the phone number of the LocationBase object."""
        return self.phone

    def set_state(self, data):
        """Set the state name of the LocationBase object."""
        self.state = data

    def get_state(self):
        """Return the state name of the LocationBase object."""
        return self.state

    def set_country(self, data):
        """Set the country name of the LocationBase object."""
        self.country = data

    def get_country(self):
        """Return the country name of the LocationBase object."""
        return self.country

    def set_county(self, data):
        """Set the county name of the LocationBase object."""
        self.county = data

    def get_county(self):
        """Return the county name of the LocationBase object."""
        return self.county



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2015-      Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# https://en.wikipedia.org/wiki/Miscellaneous_Symbols
# http://www.w3schools.com/charsets/ref_utf_symbols.asp
#

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.config import config

_ = glocale.translation.sgettext

# pylint: disable=superfluous-parens
# pylint: disable=anomalous-unicode-escape-in-string


class Symbols(object):
    # genealogical symbols
    SYMBOL_FEMALE = 0
    SYMBOL_MALE = 1
    SYMBOL_ASEXUAL_SEXLESS = 2  # Unknown
    SYMBOL_HERMAPHRODITE = 3  # Other
    SYMBOL_LESBIAN = 4
    SYMBOL_MALE_HOMOSEXUAL = 5
    SYMBOL_HETEROSEXUAL = 6
    SYMBOL_TRANSGENDER = 7
    SYMBOL_NEUTER = 8
    SYMBOL_ILLEGITIM = 9
    SYMBOL_BIRTH = 10
    SYMBOL_BAPTISM = 11  # CHRISTENING
    SYMBOL_ENGAGED = 12
    SYMBOL_MARRIAGE = 13
    SYMBOL_DIVORCE = 14
    SYMBOL_UNMARRIED_PARTNERSHIP = 15
    SYMBOL_BURIED = 16
    SYMBOL_CREMATED = 17  # Funeral urn
    SYMBOL_KILLED_IN_ACTION = 18
    SYMBOL_EXTINCT = 19

    # genealogical death symbols
    DEATH_SYMBOL_NONE = 0
    DEATH_SYMBOL_X = 1
    DEATH_SYMBOL_SKULL = 2
    DEATH_SYMBOL_ANKH = 3
    DEATH_SYMBOL_ORTHODOX_CROSS = 4
    DEATH_SYMBOL_CHI_RHO = 5
    DEATH_SYMBOL_LORRAINE_CROSS = 6
    DEATH_SYMBOL_JERUSALEM_CROSS = 7
    DEATH_SYMBOL_STAR_CRESCENT = 8
    DEATH_SYMBOL_WEST_SYRIAC_CROSS = 9
    DEATH_SYMBOL_EAST_SYRIAC_CROSS = 10
    DEATH_SYMBOL_HEAVY_GREEK_CROSS = 11
    DEATH_SYMBOL_LATIN_CROSS = 12
    DEATH_SYMBOL_SHADOWED_LATIN_CROSS = 13
    DEATH_SYMBOL_MALTESE_CROSS = 14
    DEATH_SYMBOL_STAR_OF_DAVID = 15
    DEATH_SYMBOL_DEAD = 16

    def __init__(self):
        self.symbols = None
        self.all_symbols = [
            # Name        UNICODE   SUBSTITUTION
            (_("Female"), "\u2640", ""),
            (_("Male"), "\u2642", ""),
            (_("Asexuality, sexless, genderless"), "\u26aa", ""),
            (_("Transgender, hermaphrodite (in entomology)"), "\u26a5", ""),
            (_("Lesbianism"), "\u26a2", ""),
            (_("Male homosexuality"), "\u26a3", ""),
            (_("Heterosexuality"), "\u26a4", ""),
            (_("Transgender"), "\u26a6", ""),
            (_("Neuter"), "\u26b2", ""),
            (_("Illegitimate"), "\u229b", ""),
            (_("Birth"), "\u002a", config.get("utf8.birth-symbol")),
            (_("Baptism/Christening"), "\u007e", config.get("utf8.baptism-symbol")),
            (_("Engaged"), "\u26ac", config.get("utf8.engaged-symbol")),
            (_("Marriage"), "\u26ad", config.get("utf8.marriage-symbol")),
            (_("Divorce"), "\u26ae", config.get("utf8.divorce-symbol")),
            (_("Unmarried partnership"), "\u26af", config.get("utf8.partner-symbol")),
            (_("Buried"), "\u26b0", config.get("utf8.buried-symbol")),
            (_("Cremated/Funeral urn"), "\u26b1", config.get("utf8.cremated-symbol")),
            (_("Killed in action"), "\u2694", config.get("utf8.killed-symbol")),
            (_("Extinct"), "\u2021", ""),
        ]

        # The following is used in the global preferences in the display tab.
        #                     Name       UNICODE    SUBSTITUTION
        self.death_symbols = [
            (_("Nothing"), "", ""),
            ("x", "x", "x"),
            (_("Skull and crossbones"), "\u2620", config.get("utf8.dead-symbol")),
            (_("Ankh"), "\u2625", config.get("utf8.dead-symbol")),
            (_("Orthodox cross"), "\u2626", config.get("utf8.dead-symbol")),
            (_("Chi rho"), "\u2627", config.get("utf8.dead-symbol")),
            (_("Cross of Lorraine"), "\u2628", config.get("utf8.dead-symbol")),
            (_("Cross of Jerusalem"), "\u2629", config.get("utf8.dead-symbol")),
            (_("Star and crescent"), "\u262a", config.get("utf8.dead-symbol")),
            (_("West Syriac cross"), "\u2670", config.get("utf8.dead-symbol")),
            (_("East Syriac cross"), "\u2671", config.get("utf8.dead-symbol")),
            (_("Heavy Greek cross"), "\u271a", config.get("utf8.dead-symbol")),
            (_("Latin cross"), "\u271d", config.get("utf8.dead-symbol")),
            (_("Shadowed White Latin cross"), "\u271e", config.get("utf8.dead-symbol")),
            (_("Maltese cross"), "\u2720", config.get("utf8.dead-symbol")),
            (_("Star of David"), "\u2721", config.get("utf8.dead-symbol")),
            (_("Dead"), ("Dead"), _("Dead")),
        ]

    #
    # functions for general symbols
    #
    def get_symbol_for_html(self, symbol):
        """return the html string like '&#9898;'"""
        return "&#%d;" % ord(self.all_symbols[symbol][1])

    def get_symbol_name(self, symbol):
        """
        Return the name of the symbol.
        """
        return self.all_symbols[symbol][0]

    def get_symbol_for_string(self, symbol):
        """return the utf-8 character like '\u2670'"""
        return self.all_symbols[symbol][1]

    def get_symbol_fallback(self, symbol):
        """
        Return the replacement string.
        This is used if the utf-8 symbol in not present within a font.
        """
        return self.all_symbols[symbol][2]

    #
    # functions for death symbols
    #
    def get_death_symbols(self):
        """
        Return the list of death symbols.
        This is used in the global preference to choose which symbol we'll use.
        """
        return self.death_symbols

    def get_death_symbol_name(self, symbol):
        """
        Return the name of the symbol.
        """
        return self.death_symbols[symbol][0]

    def get_death_symbol_for_html(self, symbol):
        """
        return the html string like '&#9898;'.
        """
        return "&#%d;" % ord(self.death_symbols[symbol][1])

    def get_death_symbol_for_char(self, symbol):
        """
        Return the utf-8 character for the symbol.
        """
        return self.death_symbols[symbol][1]

    def get_death_symbol_fallback(self, symbol):
        """
        Return the string replacement for the symbol.
        """
        return self.death_symbols[symbol][2]

    #
    # functions for all symbols
    #
    def get_how_many_symbols(self):
        return len(self.death_symbols) + len(self.all_symbols) - 4



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
AttributeRootBase class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .attribute import Attribute, AttributeRoot
from .srcattribute import SrcAttribute
from .const import IDENTICAL, EQUAL


# -------------------------------------------------------------------------
#
# AttributeRootBase class
#
# -------------------------------------------------------------------------
class AttributeRootBase:
    """
    Base class for attribute-aware objects.
    """

    _CLASS = AttributeRoot

    def __init__(self, source=None):
        """
        Initialize a AttributeBase.

        If the source is not None, then object is initialized from values of
        the source object.

        :param source: Object used to initialize the new object
        :type source: AttributeBase
        """
        if source:
            self.attribute_list = [
                self._CLASS(attribute) for attribute in source.attribute_list
            ]
        else:
            self.attribute_list = []

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return [attr.serialize() for attr in self.attribute_list]

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        self.attribute_list = [self._CLASS().unserialize(item) for item in data]

    def add_attribute(self, attribute):
        """
        Add the :class:`~.attribute.Attribute` instance to the object's list of
        attributes.

        :param attribute: :class:`~.attribute.Attribute` instance to add.
        :type attribute: :class:`~.attribute.Attribute`
        """
        assert not isinstance(attribute, str)
        self.attribute_list.append(attribute)

    def remove_attribute(self, attribute):
        """
        Remove the specified :class:`~.attribute.Attribute` instance from the
        attribute list.

        If the instance does not exist in the list, the operation has
        no effect.

        :param attribute: :class:`~.attribute.Attribute` instance to remove
                          from the list
        :type attribute: :class:`~.attribute.Attribute`

        :returns: True if the attribute was removed, False if it was not in the
                  list.
        :rtype: bool
        """
        if attribute in self.attribute_list:
            self.attribute_list.remove(attribute)
            return True
        else:
            return False

    def get_attribute_list(self):
        """
        Return the list of :class:`~.attribute.Attribute` instances associated
        with the object.

        :returns: Returns the list of :class:`~.attribute.Attribute` instances.
        :rtype: list
        """
        return self.attribute_list

    def set_attribute_list(self, attribute_list):
        """
        Assign the passed list to the Person's list of
        :class:`~.attribute.Attribute` instances.

        :param attribute_list: List of :class:`~.attribute.Attribute` instances
                               to ba associated with the Person
        :type attribute_list: list
        """
        self.attribute_list = attribute_list

    def _merge_attribute_list(self, acquisition):
        """
        Merge the list of attributes from acquisition with our own.

        :param acquisition: the attribute list of this object will be merged
                            with the current attribute list.
        :type acquisition: AttributeBase
        """
        attr_list = self.attribute_list[:]
        for addendum in acquisition.get_attribute_list():
            for attr in attr_list:
                equi = attr.is_equivalent(addendum)
                if equi == IDENTICAL:
                    break
                elif equi == EQUAL:
                    attr.merge(addendum)
                    break
            else:
                self.attribute_list.append(addendum)


class AttributeBase(AttributeRootBase):
    _CLASS = Attribute


class SrcAttributeBase(AttributeRootBase):
    _CLASS = SrcAttribute



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from gramps.gen.display.name import displayer as name_displayer


# -------------------------------------------------------------------------
#
# PersonRefModel
#
# -------------------------------------------------------------------------
class PersonRefModel(Gtk.ListStore):
    def __init__(self, obj_list, db):
        Gtk.ListStore.__init__(self, str, str, str, bool, bool, object)
        self.db = db
        for obj in obj_list:
            p = self.db.get_person_from_handle(obj.ref)
            if p:
                data = [
                    name_displayer.display(p),
                    p.gramps_id,
                    obj.rel,
                    obj.has_citations(),
                    obj.get_privacy(),
                    obj,
                ]
            else:
                data = ["unknown", "unknown", obj.rel, obj.get_privacy(), obj]
            self.append(row=data)



# -*- python -*-
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011-2016  Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Geography for one person and all his descendant
"""
# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import operator

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GLib

# -------------------------------------------------------------------------
#
# Gramps Modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.lib import EventRoleType, EventType
from gramps.gen.config import config
from gramps.gen.datehandler import displayer
from gramps.gen.display.name import displayer as _nd
from gramps.gen.display.place import displayer as _pd
from gramps.gen.utils.place import conv_lat_lon
from gramps.gui.views.bookmarks import PersonBookmarks
from gramps.plugins.lib.maps import constants
from gramps.plugins.lib.maps.geography import GeoGraphyView

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

_LOG = logging.getLogger("GeoGraphy.geomoves")
_ = glocale.translation.gettext

_UI_DEF = [
    """
      <placeholder id="CommonGo">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">_Forward</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="action">win.HomePerson</attribute>
          <attribute name="label" translatable="yes">_Home</attribute>
        </item>
      </section>
      </placeholder>
""",
    """
      <section id='CommonEdit' groups='RW'>
        <item>
          <attribute name="action">win.PrintView</attribute>
          <attribute name="label" translatable="yes">Print...</attribute>
        </item>
      </section>
""",
    """
      <section id="AddEditBook">
        <item>
          <attribute name="action">win.AddBook</attribute>
          <attribute name="label" translatable="yes">_Add Bookmark</attribute>
        </item>
        <item>
          <attribute name="action">win.EditBook</attribute>
          <attribute name="label" translatable="no">%s...</attribute>
        </item>
      </section>
"""
    % _("Organize Bookmarks"),  # Following are the Toolbar items
    """
    <placeholder id='CommonNavigation'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-previous</property>
        <property name="action-name">win.Back</property>
        <property name="tooltip_text" translatable="yes">"""
    """Go to the previous object in the history</property>
        <property name="label" translatable="yes">_Back</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-next</property>
        <property name="action-name">win.Forward</property>
        <property name="tooltip_text" translatable="yes">"""
    """Go to the next object in the history</property>
        <property name="label" translatable="yes">_Forward</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-home</property>
        <property name="action-name">win.HomePerson</property>
        <property name="tooltip_text" translatable="yes">"""
    """Go to the home person</property>
        <property name="label" translatable="yes">_Home</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
""",
    """
    <placeholder id='BarCommonEdit'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">document-print</property>
        <property name="action-name">win.PrintView</property>
        <property name="tooltip_text" translatable="yes">"""
    """Print or save the Map</property>
        <property name="label" translatable="yes">Print...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
    """,
]

# pylint: disable=no-member
# pylint: disable=unused-variable
# pylint: disable=unused-argument


# -------------------------------------------------------------------------
#
# GeoView : GeoMoves
#
# -------------------------------------------------------------------------
class GeoMoves(GeoGraphyView):
    """
    The view used to render all places visited by one person and all
    his descendants.
    """

    CONFIGSETTINGS = (
        ("geography.path", constants.GEOGRAPHY_PATH),
        ("geography.zoom", 10),
        ("geography.zoom_when_center", 12),
        ("geography.show_cross", True),
        ("geography.lock", True),
        ("geography.center-lat", 0.0),
        ("geography.center-lon", 0.0),
        ("geography.use-keypad", True),
        ("geography.personal-map", ""),
        ("geography.map_service", constants.OPENSTREETMAP),
        ("geography.max_places", 5000),
        # specific to geoclose :
        ("geography.color_base", "orange"),
        ("geography.maximum_generations", 10),
        ("geography.generation_interval", 500),
    )

    def __init__(self, pdata, dbstate, uistate, nav_group=0):
        GeoGraphyView.__init__(
            self,
            _("Descendants of the active person."),
            pdata,
            dbstate,
            uistate,
            PersonBookmarks,
            nav_group,
        )
        self.dbstate = dbstate
        self.uistate = uistate
        self.place_list = []
        self.place_without_coordinates = []
        self.minlat = self.maxlat = self.minlon = self.maxlon = 0.0
        self.minyear = 9999
        self.maxyear = 0
        self.started = False
        self.nbplaces = 0
        self.nbmarkers = 0
        self.sort = []
        self.tracks = []
        self.additional_uis.append(self.additional_ui())
        self.skip_list = []
        self.track = []
        self.place_list_active = []
        self.place_list_ref = []
        self.cal = config.get("preferences.calendar-format-report")
        self.markers_by_level = dict()
        self.count = dict()
        self.no_show_places_in_status_bar = False
        self.person_list = []

    def get_title(self):
        """
        Used to set the titlebar in the configuration window.
        """
        return _("GeoMoves")

    def get_stock(self):
        """
        Returns the name of the stock icon to use for the display.
        This assumes that this icon has already been registered
        as a stock icon.
        """
        return "geo-show-family-down"

    def get_viewtype_stock(self):
        """Type of view in category"""
        return "geo-show-family-down"

    def additional_ui(self):
        """
        Specifies the UIManager XML code that defines the menus and buttons
        associated with the interface.
        """
        return _UI_DEF

    def navigation_type(self):
        """
        Indicates the navigation type. Navigation type can be the string
        name of any of the primary objects.
        """
        return "Person"

    def goto_handle(self, handle=None):
        """
        Rebuild the tree with the given family handle as reference.
        """
        if self.osm is None:
            return
        if not self.started:
            self.started = True
        self.place_list_active = []
        self.place_list_ref = []
        self.sort = []
        self.remove_all_gps()
        self.remove_all_markers()
        self.lifeway_layer.clear_ways()
        self.date_layer.clear_dates()
        self.message_layer.clear_messages()
        self.message_layer.set_font_attributes(None, None, None)
        self.places_found = []
        self.place_without_coordinates = []
        active = self.get_active()
        if active:
            person = self.dbstate.db.get_person_from_handle(active)
            self._createmap(person)
        self.uistate.modify_statusbar(self.dbstate)

    def build_tree(self):
        """
        This is called by the parent class when the view becomes visible. Since
        all handling of visibility is now in rebuild_trees, see that for more
        information.
        """
        if self.osm is None:
            return
        self.place_list_active = []
        self.place_list_ref = []
        self.sort = []
        self.places_found = []
        self.place_without_coordinates = []
        self.remove_all_gps()
        self.remove_all_markers()
        self.lifeway_layer.clear_ways()
        self.date_layer.clear_dates()
        self.message_layer.clear_messages()
        self.message_layer.set_font_attributes(None, None, None)

    def draw(self, menu, marks, color):
        """
        Create all displacements for one person's events.
        """
        points = []
        mark = None
        date = "    "
        for mark in marks:
            startlat = float(mark[3])
            startlon = float(mark[4])
            not_stored = True
            for idx in range(0, len(points)):
                if points[idx][0] == startlat and points[idx][1] == startlon:
                    not_stored = False
            if not_stored:
                points.append((startlat, startlon))
                if mark[6] is not None:
                    date = mark[6]
                if date != "    ":
                    self.date_layer.add_date(date[0:4])
        self.lifeway_layer.add_way(points, color)
        return False

    def _createmap_for_one_person(self, person, color):
        """
        Create all markers for each people's event in the database which has
        a lat/lon.
        """
        if self.stop:  # no more database. stop to work
            return
        self.place_list = []
        dbstate = self.dbstate
        if person is not None:
            # For each event, if we have a place, set a marker.
            for event_ref in person.get_event_ref_list():
                if not event_ref:
                    continue
                event = dbstate.db.get_event_from_handle(event_ref.ref)
                role = event_ref.get_role()
                try:
                    date = event.get_date_object().to_calendar(self.cal)
                except:
                    continue
                eyear = (
                    str("%04d" % date.get_year())
                    + str("%02d" % date.get_month())
                    + str("%02d" % date.get_day())
                )
                place_handle = event.get_place_handle()
                if place_handle:
                    place = dbstate.db.get_place_from_handle(place_handle)
                    if place:
                        longitude = place.get_longitude()
                        latitude = place.get_latitude()
                        latitude, longitude = conv_lat_lon(latitude, longitude, "D.D8")
                        descr = _pd.display(dbstate.db, place)
                        evt = EventType(event.get_type())
                        descr1 = _("%(eventtype)s : %(name)s") % {
                            "eventtype": evt,
                            "name": _nd.display(person),
                        }
                        # place.get_longitude and place.get_latitude return
                        # one string. We have coordinates when the two values
                        # contains non null string.
                        if longitude and latitude:
                            self._append_to_places_list(
                                descr,
                                evt,
                                person.gramps_id,
                                latitude,
                                longitude,
                                descr1,
                                eyear,
                                event.get_type(),
                                person.gramps_id,
                                place.gramps_id,
                                event.gramps_id,
                                role,
                            )
                        else:
                            self._append_to_places_without_coord(place.gramps_id, descr)
            family_list = person.get_family_handle_list()
            descr1 = " - "
            for family_hdl in family_list:
                family = self.dbstate.db.get_family_from_handle(family_hdl)
                if family is not None:
                    fhandle = family_list[0]  # first is primary
                    fam = dbstate.db.get_family_from_handle(fhandle)
                    mother = father = None
                    handle = fam.get_father_handle()
                    if handle:
                        father = dbstate.db.get_person_from_handle(handle)
                    if father:
                        descr1 = "%s - " % _nd.display(father)
                    handle = fam.get_mother_handle()
                    if handle:
                        mother = dbstate.db.get_person_from_handle(handle)
                    if mother:
                        descr1 = "%s%s" % (descr1, _nd.display(mother))
                    for event_ref in family.get_event_ref_list():
                        if event_ref:
                            event = dbstate.db.get_event_from_handle(event_ref.ref)
                            role = event_ref.get_role()
                            if event.get_place_handle():
                                place_handle = event.get_place_handle()
                                if place_handle:
                                    place = dbstate.db.get_place_from_handle(
                                        place_handle
                                    )
                                    if place:
                                        longitude = place.get_longitude()
                                        latitude = place.get_latitude()
                                        latitude, longitude = conv_lat_lon(
                                            latitude, longitude, "D.D8"
                                        )
                                        descr = _pd.display(dbstate.db, place)
                                        evt = EventType(event.get_type())
                                        eyear = (
                                            str(
                                                "%04d"
                                                % event.get_date_object()
                                                .to_calendar(self.cal)
                                                .get_year()
                                            )
                                            + str(
                                                "%02d"
                                                % event.get_date_object()
                                                .to_calendar(self.cal)
                                                .get_month()
                                            )
                                            + str(
                                                "%02d"
                                                % event.get_date_object()
                                                .to_calendar(self.cal)
                                                .get_day()
                                            )
                                        )
                                        if longitude and latitude:
                                            self._append_to_places_list(
                                                descr,
                                                evt,
                                                person.gramps_id,
                                                latitude,
                                                longitude,
                                                descr1,
                                                eyear,
                                                event.get_type(),
                                                person.gramps_id,
                                                place.gramps_id,
                                                event.gramps_id,
                                                role,
                                            )
                                        else:
                                            self._append_to_places_without_coord(
                                                place.gramps_id, descr
                                            )

            sort1 = sorted(self.place_list, key=operator.itemgetter(1, 6))
            self.draw(None, sort1, color)
            # merge with the last results
            merge_list = self.sort
            for the_event in sort1:
                if the_event not in merge_list:
                    merge_list.append(the_event)
            self.sort = sorted(merge_list, key=operator.itemgetter(6))

    def _add_person_to_list(self, person_id, level):
        """
        Create a list of uniq person.
        """
        if [person_id, level] not in self.person_list:
            self.person_list.append([person_id, level])
            try:
                self.count[level] += 1
            except:
                self.count[level] = 1

    def _prepare_for_one_family(self, family, level, curlevel):
        """
        Create all markers for one family : all event's places with a lat/lon.
        """
        if self.stop:  # no more database. stop to work
            return
        dbstate = self.dbstate
        person = None
        try:
            person = dbstate.db.get_person_from_handle(family.get_father_handle())
        except:
            return
        family_id = family.gramps_id
        if person is None:  # family without father ?
            handle = family.get_mother_handle()
            if handle:
                person = dbstate.db.get_person_from_handle(handle)
        if person is None:
            handle = self.uistate.get_active("Person")
            if handle:
                person = dbstate.db.get_person_from_handle(handle)
        if person is not None:
            self._add_person_to_list(person.gramps_id, curlevel - 1)
            family_list = person.get_family_handle_list()
            for fhandle in family_list:
                fam = dbstate.db.get_family_from_handle(fhandle)
                father = mother = None
                handle = fam.get_father_handle()
                if handle:
                    father = dbstate.db.get_person_from_handle(handle)
                if father:
                    self._createmap_for_next_level(father, level - 1, level)
                    self._add_person_to_list(father.gramps_id, curlevel - 1)
                handle = fam.get_mother_handle()
                if handle:
                    mother = dbstate.db.get_person_from_handle(handle)
                if mother:
                    self._createmap_for_next_level(father, level - 1, level)
                    self._add_person_to_list(mother.gramps_id, curlevel - 1)
                index = 0
                child_ref_list = fam.get_child_ref_list()
                if child_ref_list:
                    for child_ref in child_ref_list:
                        child = dbstate.db.get_person_from_handle(child_ref.ref)
                        if child:
                            index += 1
                            self._createmap_for_next_level(child, level, curlevel)
                            self._add_person_to_list(child.gramps_id, curlevel)

    def _createmap_for_one_level(self, family, level, curlevel):
        """
        if maximum generation is not reached, show next level.
        """
        if self.stop:  # no more database. stop to work
            return
        if level < curlevel:
            return
        self._prepare_for_one_family(family, level, curlevel + 1)

    def _createmap_for_next_level(self, person, level, curlevel):
        """
        if maximum generation is not reached, show next level.
        """
        if level < curlevel:
            return
        try:
            if person not in self.markers_by_level[curlevel]:
                self.markers_by_level[curlevel].append(person)
        except:
            self.markers_by_level[curlevel] = []
            self.markers_by_level[curlevel].append(person)
        if self.stop:  # no more database. stop to work
            return
        for family in person.get_family_handle_list():
            if self.stop:  # no more database. stop to work
                return
            fam = self.dbstate.db.get_family_from_handle(family)
            self._createmap_for_one_level(fam, level, curlevel)
            if person not in self.markers_by_level[curlevel]:
                self.markers_by_level[curlevel].append(person)

    def _createmap(self, person):
        """
        Create all markers for each family's person in the database which has
        a lat/lon.
        """
        dbstate = self.dbstate
        self.cal = config.get("preferences.calendar-format-report")
        self.place_list = []
        self.person_list = []
        self.count = dict()
        self.place_without_coordinates = []
        self.minlat = self.maxlat = self.minlon = self.maxlon = 0.0
        latitude = ""
        longitude = ""
        self.places_found = []
        self.nbplaces = 0
        self.nbmarkers = 0
        self.message_layer.clear_messages()
        self.place_without_coordinates = []
        self.minlat = self.maxlat = self.minlon = self.maxlon = 0.0
        if person is None:
            handle = self.uistate.get_active("Person")
            if handle:
                person = self.dbstate.db.get_person_from_handle(handle)
            if not person:
                return
        self.message_layer.add_message(
            _("All descendants for %s") % _nd.display(person)
        )
        color = Gdk.color_parse(self._config.get("geography.color_base"))
        GLib.timeout_add(
            int(self._config.get("geography.generation_interval")),
            self.animate_moves,
            0,
            person,
            color,
        )

    def animate_moves(self, index, person, color):
        """
        Animate all moves for one generation.
        """
        if self.stop:  # no more database. stop to work
            return
        self.markers_by_level = dict()
        self._createmap_for_next_level(person, index, 0)
        try:
            persons = self.markers_by_level[index]
        except:
            return
        for people in persons:
            family_list = people.get_family_handle_list()
            for fhandle in family_list:
                family = self.dbstate.db.get_family_from_handle(fhandle)
                self._prepare_for_one_family(family, index, index + 1)
        new_list = []
        for plx, level in self.person_list:
            plxp = self.dbstate.db.get_person_from_gramps_id(plx)
            birth = "0000"
            death = "0000"
            low_date = "9999"
            high_date = "0000"
            for event_ref in plxp.get_event_ref_list():
                if not event_ref:
                    continue
                event = self.dbstate.db.get_event_from_handle(event_ref.ref)
                role = event_ref.get_role()
                try:
                    date = event.get_date_object().to_calendar(self.cal)
                    fyear = str("%04d" % date.get_year())
                    if event.get_type() == EventType.BIRTH:
                        birth = fyear
                    if event.get_type() == EventType.DEATH:
                        death = fyear
                    if fyear < low_date:
                        low_date = fyear
                    if fyear > high_date:
                        high_date = fyear

                except:
                    pass
                if birth == "0000":
                    birth = low_date
                if death == "0000":
                    death = high_date
            new_list.append([level, plxp, birth, death])
        pidx = 0
        if isinstance(color, str):
            color = Gdk.color_parse(color)
        for level, plxp, birth, death in sorted(
            new_list, key=operator.itemgetter(0, 2)
        ):
            if index == int(self._config.get("geography.maximum_generations")):
                break
            if level == index:
                pidx += 1
                self._createmap_for_one_person(plxp, color)
                color.red = float(color.red - (index) * 3000) % 65535
                if index % 2:
                    color.green = float((color.green + (index) * 3000) % 65535)
                else:
                    color.blue = float((color.blue + (index) * 3000) % 65535)
                self._createmap_for_one_person(person, color)
        if index < int(self._config.get("geography.maximum_generations")):
            time_to_wait = int(self._config.get("geography.generation_interval"))
            self._create_markers()
            # process next generation in a few milliseconds
            GLib.timeout_add(
                int(time_to_wait), self.animate_moves, index + 1, person, color
            )
        else:
            self.started = False
        return False

    def bubble_message(self, event, lat, lon, marks):
        """
        Create the menu for the selected marker
        """
        self.menu = Gtk.Menu()
        menu = self.menu
        events = []
        message = ""
        oldplace = ""
        prevmark = None
        # Be sure all markers are sorted by place then dates.
        for mark in sorted(marks, key=operator.itemgetter(0, 6)):
            if mark[10] in events:
                continue  # avoid duplicate events
            else:
                events.append(mark[10])
            if mark[0] != oldplace:
                message = "%s :" % mark[0]
                self.add_place_bubble_message(
                    event, lat, lon, marks, menu, message, mark
                )
                oldplace = mark[0]
                message = ""
            evt = self.dbstate.db.get_event_from_gramps_id(mark[10])
            # format the date as described in preferences.
            date = displayer.display(evt.get_date_object())
            if date == "":
                date = _("Unknown")
            if mark[11] == EventRoleType.PRIMARY:
                person = self.dbstate.db.get_person_from_gramps_id(mark[1])
                message = "(%s) %s : %s" % (date, mark[2], _nd.display(person))
            elif mark[11] == EventRoleType.FAMILY:
                (father_name, mother_name) = self._get_father_and_mother_name(evt)
                message = "(%s) %s : %s - %s" % (
                    date,
                    mark[2],
                    father_name,
                    mother_name,
                )
            else:
                descr = evt.get_description()
                if descr == "":
                    descr = _("No description")
                message = "(%s) %s => %s" % (date, mark[11], descr)
            prevmark = mark
            add_item = Gtk.MenuItem(label=message)
            add_item.show()
            menu.append(add_item)
            self.itemoption = Gtk.Menu()
            itemoption = self.itemoption
            itemoption.show()
            add_item.set_submenu(itemoption)
            modify = Gtk.MenuItem(label=_("Edit Event"))
            modify.show()
            modify.connect("activate", self.edit_event, event, lat, lon, prevmark)
            itemoption.append(modify)
            center = Gtk.MenuItem(label=_("Center on this place"))
            center.show()
            center.connect("activate", self.center_here, event, lat, lon, prevmark)
            itemoption.append(center)
            person = self.dbstate.db.get_person_from_gramps_id(mark[8])
            hdle = person.get_handle()
            bookm = Gtk.MenuItem(label=_("Bookmark this person"))
            bookm.show()
            bookm.connect("activate", self.add_bookmark_from_popup, hdle)
            itemoption.append(bookm)
            menu.show()
            menu.popup_at_pointer(event)
        return 1

    def add_specific_menu(self, menu, event, lat, lon):
        """
        Add specific entry to the navigation menu.
        """
        return

    def get_default_gramplets(self):
        """
        Define the default gramplets for the sidebar and bottombar.
        """
        return (("Person Filter",), ())

    def specific_options(self, configdialog):
        """
        Add specific entry to the preference menu.
        Must be done in the associated view.
        """
        grid = Gtk.Grid()
        grid.set_border_width(12)
        grid.set_column_spacing(6)
        grid.set_row_spacing(6)
        configdialog.add_text(
            grid, _("The maximum number of generations.\n"), 1, line_wrap=False
        )
        configdialog.add_slider(grid, "", 2, "geography.maximum_generations", (1, 20))
        configdialog.add_text(
            grid,
            _("Time in milliseconds between drawing two" " generations.\n"),
            3,
            line_wrap=False,
        )
        configdialog.add_slider(
            grid, "", 4, "geography.generation_interval", (500, 3000)
        )
        return _("The parameters for moves"), grid

    def config_connect(self):
        """
        used to monitor changes in the ini file
        """
        self._config.connect("geography.maximum_generations", self._maximum_generations)

    def _maximum_generations(self, client, cnxn_id, entry, data):
        """
        Called when the number of nomber of generations change
        """
        self.goto_handle(handle=None)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from collections import defaultdict

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.const import PLUGINS_GLADE
from gramps.gen.plug.report._constants import standalone_categories
from . import tool
from gramps.gen.plug import REPORT
from .report import report
from ..pluginmanager import GuiPluginManager
from ..managedwindow import ManagedWindow

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
_REPORTS = 0
_TOOLS = 1
_UNSUPPORTED = _("Unsupported")


# -------------------------------------------------------------------------
#
# PluginDialog interface class
#
# -------------------------------------------------------------------------
class PluginDialog(ManagedWindow):
    """
    Displays the dialog box that allows the user to select the
    plugin that is desired.
    """

    def __init__(
        self,
        state,
        uistate,
        track,
        categories,
        msg,
        label=None,
        button_label=None,
        tool_tip=None,
        content=_REPORTS,
    ):
        """
        Display the dialog box, and build up the list of available
        reports. This is used to build the selection tree on the left
        hand side of the dialog box.
        """
        self.active = uistate.get_active("Person")
        self.imap = {}
        self.msg = msg
        self.content = content
        self._pmgr = GuiPluginManager.get_instance()

        ManagedWindow.__init__(self, uistate, track, self.__class__)

        self.state = state
        self.uistate = uistate

        self.dialog = Gtk.Builder()
        self.dialog.set_translation_domain(glocale.get_localedomain())
        self.dialog.add_from_file(PLUGINS_GLADE)
        self.dialog.connect_signals(
            {
                "on_report_apply_clicked": self.on_apply_clicked,
                "destroy_passed_object": self.close,
                "on_delete_event": self.close,
            }
        )

        self.tree = self.dialog.get_object("tree")
        window = self.dialog.get_object("report")
        self.title = self.dialog.get_object("title")

        self.set_window(window, self.title, msg)
        if content == _REPORTS:
            self.setup_configs("interface.reportplugindialog", 600, 400)
        elif content == _TOOLS:
            self.setup_configs("interface.toolplugindialog", 600, 400)
        else:
            raise ValueError("PluginDialog selection must be a report or tool")

        self.store = Gtk.TreeStore(str)
        self.selection = self.tree.get_selection()
        self.selection.connect("changed", self.on_node_selected)
        col = Gtk.TreeViewColumn("", Gtk.CellRendererText(), text=0)
        self.tree.append_column(col)
        self.tree.set_model(self.store)

        self.description = self.dialog.get_object("description")
        if label:
            self.description.set_text(label)
        self.status = self.dialog.get_object("report_status")

        self.author_name = self.dialog.get_object("author_name")
        self.author_email = self.dialog.get_object("author_email")

        self.apply_button = self.dialog.get_object("apply")
        if button_label:
            self.apply_button.set_label(button_label)
        else:
            self.apply_button.set_label(_("_Apply"))
        self.apply_button.set_use_underline(True)
        if tool_tip:
            self.apply_button.set_tooltip_text(tool_tip)

        self.item = None

        if content == _REPORTS:
            reg_list = self._pmgr.get_reg_reports()
        elif content == _TOOLS:
            reg_list = self._pmgr.get_reg_tools()
        else:
            reg_list = []
        self.build_plugin_tree(reg_list, categories)
        self.show()

    def rebuild(self):
        # This method needs to be overridden in the subclass
        assert False, "This method needs to be overridden in the subclass."

    def build_menu_names(self, obj):
        return (self.msg, None)

    def on_apply_clicked(self, obj):
        """Execute the selected report"""
        if not self.item:
            return
        self.run_plugin(self.item)

    def on_node_selected(self, obj):
        """Updates the informational display on the right hand side of
        the dialog box with the description of the selected report"""

        store, node = self.selection.get_selected()
        if node:
            path = store.get_path(node).to_string()
        if not node or path not in self.imap:
            return
        pdata = self.imap[path]

        # (report_class, options_class, title, category, name,
        # doc,status,author,email,unsupported,require_active) = data
        self.description.set_text(pdata.description)
        if not pdata.supported:
            status = _UNSUPPORTED
        self.status.set_text(pdata.statustext())
        self.title.set_text('<span weight="bold" size="larger">%s</span>' % pdata.name)
        self.title.set_use_markup(1)
        # TODO for Arabic, should the next two lines' commas be translated?
        self.author_name.set_text(", ".join(pdata.authors))
        self.author_email.set_text(", ".join(pdata.authors_email))
        self.item = pdata

    def build_plugin_tree(self, reg_list, categories):
        """Populates a GtkTree with each menu item associated with a entry
        in the lists. The list consists of PluginData objects for reports or
        tools.

        old data was (item_class, options_class,title,category, name,
         doc,status,author,email)

        Items in the same category are grouped under the same submenu.
        The categories must be dicts from integer to string.
        """
        ilist = []
        self.store.clear()

        # build the tree items and group together based on the category name
        item_hash = defaultdict(list)
        for plugin in reg_list:
            if not plugin.supported:
                category = _UNSUPPORTED
            else:
                category = categories[plugin.category]
            item_hash[category].append(plugin)

        # add a submenu for each category, and populate it with the
        # GtkTreeItems that are associated with it.
        key_list = [item for item in item_hash if item != _UNSUPPORTED]
        key_list.sort(reverse=True)

        prev = None
        if _UNSUPPORTED in item_hash:
            key = _UNSUPPORTED
            data = item_hash[key]
            node = self.store.insert_after(None, prev)
            self.store.set(node, 0, key)
            next = None
            data.sort(key=lambda x: x.name)
            for item in data:
                next = self.store.insert_after(node, next)
                ilist.append((next, item))
                self.store.set(next, 0, item.name)
        for key in key_list:
            data = item_hash[key]
            node = self.store.insert_after(None, prev)
            self.store.set(node, 0, key[1])
            next = None
            data.sort(key=lambda k: k.name)
            for item in data:
                next = self.store.insert_after(node, next)
                ilist.append((next, item))
                self.store.set(next, 0, item.name)
        for next, tab in ilist:
            path = self.store.get_path(next).to_string()
            self.imap[path] = tab

    def run_plugin(self, pdata):
        """
        run a plugin based on it's PluginData:
          1/ load plugin.
          2/ the report is run
        """
        mod = self._pmgr.load_plugin(pdata)
        if not mod:
            # import of plugin failed
            return

        if pdata.ptype == REPORT:
            report(
                self.state,
                self.uistate,
                self.uistate.get_active("Person"),
                getattr(mod, pdata.reportclass),
                getattr(mod, pdata.optionclass),
                pdata.name,
                pdata.id,
                pdata.category,
                pdata.require_active,
            )
        else:
            from ..user import User

            tool.gui_tool(
                dbstate=self.state,
                user=User(uistate=self.uistate),
                tool_class=getattr(mod, pdata.toolclass),
                options_class=getattr(mod, pdata.optionclass),
                translated_name=pdata.name,
                name=pdata.id,
                category=pdata.category,
                callback=self.state.db.request_rebuild,
            )


# -------------------------------------------------------------------------
#
# ReportPluginDialog
#
# -------------------------------------------------------------------------
class ReportPluginDialog(PluginDialog):
    """
    Displays the dialog box that allows the user to select the
    report that is desired.
    """

    def __init__(self, dbstate, uistate, track):
        """Display the dialog box, and build up the list of available
        reports. This is used to build the selection tree on the left
        hand side of the dailog box."""

        PluginDialog.__init__(
            self,
            dbstate,
            uistate,
            track,
            standalone_categories,
            _("Report Selection"),
            _("Select a report from those available on the left."),
            _("_Generate"),
            _("Generate selected report"),
            _REPORTS,
        )

        self._pmgr.connect("plugins-reloaded", self.rebuild)

    def rebuild(self):
        report_list = self._pmgr.get_reg_reports()
        self.build_plugin_tree(report_list, standalone_categories)


# -------------------------------------------------------------------------
#
# ToolPluginDialog
#
# -------------------------------------------------------------------------
class ToolPluginDialog(PluginDialog):
    """Displays the dialog box that allows the user to select the tool
    that is desired."""

    def __init__(self, dbstate, uistate, track):
        """Display the dialog box, and build up the list of available
        reports. This is used to build the selection tree on the left
        hand side of the dailog box."""

        PluginDialog.__init__(
            self,
            dbstate,
            uistate,
            track,
            tool.tool_categories,
            _("Tool Selection"),
            _("Select a tool from those available on the left."),
            _("_Run"),
            _("Run selected tool"),
            _TOOLS,
        )

    def rebuild(self):
        tool_list = self._pmgr.get_reg_tools()
        self.build_plugin_tree(tool_list, tool.tool_categories)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013-2015  Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Location utility functions
"""
from ..lib.date import Date, Today


# -------------------------------------------------------------------------
#
# get_location_list
#
# -------------------------------------------------------------------------
def get_location_list(db, place, date=None, lang=""):
    """
    Return a list of place names for display.
    """
    if date is None:
        date = __get_latest_date(place)
    visited = [place.handle]
    lines = [(__get_name(place, date, lang), place.get_type())]
    while True:
        handle = None
        for placeref in place.get_placeref_list():
            ref_date = placeref.get_date_object()
            if ref_date.is_empty() or date.match_exact(ref_date):
                handle = placeref.ref
                break
        if handle is None or handle in visited:
            break
        place = db.get_place_from_handle(handle)
        if place is None:
            break
        visited.append(handle)
        lines.append((__get_name(place, date, lang), place.get_type()))
    return lines


def __get_name(place, date, lang):
    endonym = None
    for place_name in place.get_all_names():
        name_date = place_name.get_date_object()
        if name_date.is_empty() or date.match_exact(name_date):
            if place_name.get_language() == lang:
                return place_name.get_value()
            if endonym is None:
                endonym = place_name.get_value()
    return endonym if endonym is not None else "?"


def __get_latest_date(place):
    latest_date = None
    for place_name in place.get_all_names():
        date = place_name.get_date_object()
        if date.is_empty() or date.modifier in (Date.MOD_FROM, Date.MOD_AFTER):
            return Today()
        else:
            if date.is_compound():
                date1, date2 = date.get_start_stop_range()
                date = Date(*date2)
            if date.modifier in (Date.MOD_TO, Date.MOD_BEFORE):
                date = date - 1
            if latest_date is None or date > latest_date:
                latest_date = date
    return latest_date


# -------------------------------------------------------------------------
#
# get_main_location
#
# -------------------------------------------------------------------------
def get_main_location(db, place, date=None):
    """
    Find all places in the hierarchy above the given place, and return the
    result as a dictionary of place types and names.
    """
    return dict(
        [
            (int(place_type), name)
            for name, place_type in get_location_list(db, place, date)
            if not place_type.is_custom()
        ]
    )


# -------------------------------------------------------------------------
#
# get_locations
#
# -------------------------------------------------------------------------
def get_locations(db, place):
    """
    Determine each possible route up the place hierarchy, and return a list
    containing dictionaries of place types and names.
    """
    locations = []
    todo = [(place, [(int(place.get_type()), __get_all_names(place))], [place.handle])]
    while len(todo):
        place, tree, visited = todo.pop()
        for parent in place.get_placeref_list():
            if parent.ref not in visited:
                parent_place = db.get_place_from_handle(parent.ref)
                if parent_place is not None:
                    parent_tree = tree + [
                        (int(parent_place.get_type()), __get_all_names(parent_place))
                    ]
                    parent_visited = visited + [parent.ref]
                    todo.append((parent_place, parent_tree, parent_visited))
        if len(place.get_placeref_list()) == 0:
            locations.append(dict(tree))
    return locations


def __get_all_names(place):
    return [name.get_value() for name in place.get_all_names()]


# -------------------------------------------------------------------------
#
# located_in
#
# -------------------------------------------------------------------------
def located_in(db, handle1, handle2):
    """
    Determine if the place identified by handle1 is located within the place
    identified by handle2.
    """
    place = db.get_place_from_handle(handle1)
    todo = [(place, [handle1])]
    while len(todo):
        place, visited = todo.pop()
        for parent in place.get_placeref_list():
            if parent.ref == handle2:
                return True
            if parent.ref not in visited:
                parent_place = db.get_place_from_handle(parent.ref)
                if parent_place is not None:
                    parent_visited = visited + [parent.ref]
                    todo.append((parent_place, parent_visited))
    return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#               2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
The EditAttribute module provides the AttributeEditor class. This provides a
mechanism for the user to edit attribute information.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from .editsecondary import EditSecondary
from gramps.gen.lib import NoteType
from ..glade import Glade
from .displaytabs import CitationEmbedList, NoteTab
from ..widgets import MonitoredEntry, PrivacyButton, MonitoredDataType
from gramps.gen.const import URL_MANUAL_SECT3

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

WIKI_HELP_PAGE = URL_MANUAL_SECT3
WIKI_HELP_SEC = _("Attribute_Editor_dialog", "manual")


# -------------------------------------------------------------------------
#
# EditAttribute class
#
# -------------------------------------------------------------------------
class EditAttributeRoot(EditSecondary):
    """
    Root baseclass for the root Attribute data editor
    """

    def __init__(self, state, uistate, track, attrib, title, data_list, callback):
        """
        Displays the dialog box.

        state - dbstate
        uistate - the uistate
        track - list of parent windows
        attrib - The attribute that is to be edited
        title - The title of the dialog box
        data_list - list of options for the pop down menu to help selection of a
               attribute type
        """
        self.alist = data_list
        EditSecondary.__init__(self, state, uistate, track, attrib, callback)

    def _local_init(self):
        self.top = Glade("editattribute.glade")

        self.set_window(
            self.top.toplevel, self.top.get_object("title"), _("Attribute Editor")
        )
        self.setup_configs("interface.attribute", 600, 350)

    def _connect_signals(self):
        self.define_cancel_button(self.top.get_object("cancel"))
        self.define_ok_button(self.top.get_object("ok"), self.save)
        self.define_help_button(
            self.top.get_object("help"), WIKI_HELP_PAGE, WIKI_HELP_SEC
        )

    def _setup_fields(self):
        self.value_field = MonitoredEntry(
            self.top.get_object("attr_value"),
            self.obj.set_value,
            self.obj.get_value,
            self.db.readonly,
        )

        self.priv = PrivacyButton(
            self.top.get_object("private"), self.obj, self.db.readonly
        )

        self.type_selector = MonitoredDataType(
            self.top.get_object("attr_menu"),
            self.obj.set_type,
            self.obj.get_type,
            self.db.readonly,
            custom_values=self.alist,
        )

    def _create_tabbed_pages(self):
        notebook = Gtk.Notebook()

        self._setup_notebook_tabs(notebook)
        notebook.show_all()
        self.top.get_object("vbox").pack_start(notebook, True, True, 0)

    def build_menu_names(self, attrib):
        if not attrib:
            label = _("New Attribute")
        else:
            label = str(attrib.get_type())
        if not label.strip():
            label = _("New Attribute")
        label = _("%(str1)s: %(str2)s") % {"str1": _("Attribute"), "str2": label}
        return (label, _("Attribute Editor"))

    def save(self, *obj):
        """
        Called when the OK button is pressed. Gets data from the
        form and updates the Attribute data structure.
        """
        t = self.obj.get_type()

        if t.is_custom() and str(t) == "":
            from ..dialog import ErrorDialog

            ErrorDialog(
                _("Cannot save attribute"),
                _("The attribute type cannot be empty"),
                parent=self.window,
            )
            return
        if self.callback:
            self.callback(self.obj)
        self.close()


# -------------------------------------------------------------------------
#
# EditSrcAttribute class
#
# -------------------------------------------------------------------------
class EditSrcAttribute(EditAttributeRoot):
    """
    Source attribute are minimal attributes. This Displays the editor to
    edit these.
    """

    def __init__(self, state, uistate, track, attrib, title, data_list, callback):
        """
        Displays the dialog box.

        state - dbstate
        uistate - the uistate
        track - list of parent windows
        attrib - The attribute that is to be edited
        title - The title of the dialog box
        data_list - list of options for the pop down menu to help selection of a
               attribute type
        """
        EditAttributeRoot.__init__(
            self, state, uistate, track, attrib, title, data_list, callback
        )


# -------------------------------------------------------------------------
#
# EditAttribute class
#
# -------------------------------------------------------------------------
class EditAttribute(EditAttributeRoot):
    """
    Displays a dialog that allows the user to edit an attribute.
    """

    def __init__(self, state, uistate, track, attrib, title, data_list, callback):
        """
        Displays the dialog box.

        state - dbstate
        uistate - the uistate
        track - list of parent windows
        attrib - The attribute that is to be edited
        title - The title of the dialog box
        data_list - list of options for the pop down menu to help selection of a
               attribute type
        """
        EditAttributeRoot.__init__(
            self, state, uistate, track, attrib, title, data_list, callback
        )

    def _create_tabbed_pages(self):
        notebook = Gtk.Notebook()
        self.srcref_list = CitationEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_citation_list(),
            "attribute_editor_citations",
        )
        self._add_tab(notebook, self.srcref_list)
        self.track_ref_for_deletion("srcref_list")

        self.note_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_note_list(),
            "attribute_editor_notes",
            notetype=NoteType.ATTRIBUTE,
        )
        self._add_tab(notebook, self.note_tab)
        self.track_ref_for_deletion("note_tab")

        self._setup_notebook_tabs(notebook)
        notebook.show_all()
        self.top.get_object("vbox").pack_start(notebook, True, True, 0)


# -------------------------------------------------------------------------
#
# EditFamilyAttribute class
#
# -------------------------------------------------------------------------
class EditFamilyAttribute(EditAttribute):
    """
    Displays a dialog that allows the user to edit an attribute.
    """

    def __init__(self, state, uistate, track, attrib, title, data_list, callback):
        """
        Displays the dialog box.

        parent - The class that called the Address editor.
        attrib - The attribute that is to be edited
        title - The title of the dialog box
        list - list of options for the pop down menu
        """
        EditAttribute.__init__(
            self, state, uistate, track, attrib, title, data_list, callback
        )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2010       Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import os

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import Gdk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from ..utils import open_file_with_default_application
from gramps.gen.lib import Media, NoteType
from gramps.gen.db import DbTxn
from gramps.gen.mime import get_description, get_type
from gramps.gen.utils.thumbnails import get_thumbnail_image, find_mime_type_pixbuf
from gramps.gen.utils.file import media_path_full, find_file, create_checksum
from .editprimary import EditPrimary
from ..widgets import MonitoredDate, MonitoredEntry, PrivacyButton, MonitoredTagList
from .displaytabs import (
    CitationEmbedList,
    MediaAttrEmbedList,
    NoteTab,
    MediaBackRefList,
)
from .addmedia import AddMedia
from ..dialog import ErrorDialog
from ..glade import Glade
from gramps.gen.const import URL_MANUAL_SECT2

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

WIKI_HELP_PAGE = URL_MANUAL_SECT2
WIKI_HELP_SEC = _("New_Media_dialog", "manual")


# -------------------------------------------------------------------------
#
# EditMedia
#
# -------------------------------------------------------------------------
class EditMedia(EditPrimary):
    def __init__(self, dbstate, uistate, track, obj, callback=None):
        EditPrimary.__init__(
            self,
            dbstate,
            uistate,
            track,
            obj,
            dbstate.db.get_media_from_handle,
            dbstate.db.get_media_from_gramps_id,
            callback,
        )
        if not self.obj.get_handle():
            # show the addmedia dialog immediately, with track of parent.
            AddMedia(dbstate, self.uistate, self.track, self.obj, self._update_addmedia)

    def empty_object(self):
        return Media()

    def get_menu_title(self):
        if self.obj.get_handle():
            name = self.obj.get_description()
            if not name:
                name = self.obj.get_path()
            if not name:
                name = self.obj.get_mime_type()
            if not name:
                name = _("Note")
            dialog_title = _("Media: %s") % name
        else:
            dialog_title = _("New Media")
        return dialog_title

    def _local_init(self):
        assert self.obj
        self.glade = Glade()
        self.set_window(self.glade.toplevel, None, self.get_menu_title())
        self.setup_configs("interface.media", 650, 450)

    def _connect_signals(self):
        self.define_cancel_button(self.glade.get_object("button91"))
        self.define_ok_button(self.glade.get_object("ok"), self.save)
        # TODO help button (rename glade button name)
        self.define_help_button(
            self.glade.get_object("button102"), WIKI_HELP_PAGE, WIKI_HELP_SEC
        )

    def _connect_db_signals(self):
        """
        Connect any signals that need to be connected.
        Called by the init routine of the base class (_EditPrimary).
        """
        self._add_db_signal("media-rebuild", self._do_close)
        self._add_db_signal("media-delete", self.check_for_close)

    def _setup_fields(self):
        self.date_field = MonitoredDate(
            self.glade.get_object("date_entry"),
            self.glade.get_object("date_edit"),
            self.obj.get_date_object(),
            self.uistate,
            self.track,
            self.db.readonly,
        )

        self.descr_window = MonitoredEntry(
            self.glade.get_object("description"),
            self.obj.set_description,
            self.obj.get_description,
            self.db.readonly,
        )

        self.gid = MonitoredEntry(
            self.glade.get_object("gid"),
            self.obj.set_gramps_id,
            self.obj.get_gramps_id,
            self.db.readonly,
        )

        self.tags = MonitoredTagList(
            self.glade.get_object("tag_label"),
            self.glade.get_object("tag_button"),
            self.obj.set_tag_list,
            self.obj.get_tag_list,
            self.db,
            self.uistate,
            self.track,
            self.db.readonly,
        )

        self.privacy = PrivacyButton(
            self.glade.get_object("private"), self.obj, self.db.readonly
        )

        self.pixmap = self.glade.get_object("pixmap")
        ebox = self.glade.get_object("eventbox")
        ebox.connect("button-press-event", self.button_press_event)

        self.mimetext = self.glade.get_object("type")
        self.setup_filepath()
        self.determine_mime()
        self.draw_preview()

    def determine_mime(self):
        descr = get_description(self.obj.get_mime_type())
        if descr:
            self.mimetext.set_text(descr)

        path = self.file_path.get_text()
        path_full = media_path_full(self.db, path)
        if path != self.obj.get_path() and path_full != self.obj.get_path():
            # redetermine mime
            mime = get_type(find_file(path_full))
            self.obj.set_mime_type(mime)
            descr = get_description(mime)
            if descr:
                self.mimetext.set_text(descr)
            else:
                self.mimetext.set_text(_("Unknown"))
        # if mime type not set, is note
        if not self.obj.get_mime_type():
            self.mimetext.set_text(_("Note"))

    def draw_preview(self):
        mtype = self.obj.get_mime_type()
        if mtype:
            pb = get_thumbnail_image(
                media_path_full(self.db, self.obj.get_path()), mtype
            )
            self.pixmap.set_from_pixbuf(pb)
        else:
            pb = find_mime_type_pixbuf("text/plain")
            self.pixmap.set_from_pixbuf(pb)

    def setup_filepath(self):
        self.select = self.glade.get_object("file_select")
        self.file_path = self.glade.get_object("path")

        fname = self.obj.get_path()
        self.file_path.set_text(fname)
        self.select.connect("clicked", self.select_file)

    def _create_tabbed_pages(self):
        notebook = Gtk.Notebook()

        self.citation_tab = CitationEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_citation_list(),
            "media_editor_citations",
            self.get_menu_title(),
        )
        self._add_tab(notebook, self.citation_tab)
        self.track_ref_for_deletion("citation_tab")

        self.attr_tab = MediaAttrEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_attribute_list(),
            "media_editor_attributes",
        )
        self._add_tab(notebook, self.attr_tab)
        self.track_ref_for_deletion("attr_tab")

        self.note_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_note_list(),
            "media_editor_notes",
            notetype=NoteType.MEDIA,
        )
        self._add_tab(notebook, self.note_tab)
        self.track_ref_for_deletion("note_tab")

        self.backref_tab = MediaBackRefList(
            self.dbstate,
            self.uistate,
            self.track,
            self.db.find_backlink_handles(self.obj.handle),
            "media_editor_references",
        )
        self.backref_list = self._add_tab(notebook, self.backref_tab)
        self.track_ref_for_deletion("backref_tab")
        self.track_ref_for_deletion("backref_list")

        self._setup_notebook_tabs(notebook)
        notebook.show_all()
        self.glade.get_object("vbox").pack_start(notebook, True, True, 0)

    def build_menu_names(self, person):
        return (_("Edit Media Object"), self.get_menu_title())

    def button_press_event(self, obj, event):
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            self.view_media(obj)

    def view_media(self, obj):
        if self.obj.handle:
            ref_obj = self.dbstate.db.get_media_from_handle(self.obj.handle)

            if ref_obj:
                media_path = media_path_full(self.dbstate.db, ref_obj.get_path())
                open_file_with_default_application(media_path, self.uistate)

    def select_file(self, val):
        self.determine_mime()
        path = self.file_path.get_text()
        self.obj.set_path(path)
        AddMedia(
            self.dbstate, self.uistate, self.track, self.obj, self._update_addmedia
        )

    def _update_addmedia(self, obj):
        """
        Called when the add media dialog has been called.
        This allows us to update the main form in response to
        any changes: Redraw relevant fields: description, mimetype and path
        """
        for obj in (self.descr_window,):
            obj.update()
        fname = self.obj.get_path()
        self.file_path.set_text(fname)
        self.determine_mime()
        self.update_checksum()
        self.draw_preview()

    def update_checksum(self):
        self.uistate.set_busy_cursor(True)
        media_path = media_path_full(self.dbstate.db, self.obj.get_path())
        self.obj.set_checksum(create_checksum(os.path.normpath(media_path)))
        self.uistate.set_busy_cursor(False)

    def save(self, *obj):
        self.ok_button.set_sensitive(False)

        if self.object_is_empty():
            ErrorDialog(
                _("Cannot save media object"),
                _(
                    "No data exists for this media object. Please "
                    "enter data or cancel the edit."
                ),
                parent=self.window,
            )
            self.ok_button.set_sensitive(True)
            return

        (uses_dupe_id, id) = self._uses_duplicate_id()
        if uses_dupe_id:
            prim_object = self.get_from_gramps_id(id)
            name = prim_object.get_description()
            msg1 = _("Cannot save media object. ID already exists.")
            msg2 = _(
                "You have attempted to use the existing Gramps ID with "
                "value %(id)s. This value is already used by '"
                "%(prim_object)s'. Please enter a different ID or leave "
                "blank to get the next available ID value."
            ) % {"id": id, "prim_object": name}
            ErrorDialog(msg1, msg2, parent=self.window)
            self.ok_button.set_sensitive(True)
            return

        path = self.file_path.get_text()
        full_path = media_path_full(self.db, path)
        if os.path.isfile(full_path):
            self.determine_mime()
        else:
            msg1 = _("There is no media matching the current path value!")
            msg2 = _(
                "You have attempted to use the path with "
                "value '%(path)s'. This path does not exist!"
                " Please enter a different path"
            ) % {"path": path}
            ErrorDialog(msg1, msg2, parent=self.window)
            self.ok_button.set_sensitive(True)
            return

        self.obj.set_path(path)

        if not self.obj.handle:
            with DbTxn(
                _("Add Media Object (%s)") % self.obj.get_description(), self.db
            ) as trans:
                self.db.add_media(self.obj, trans)
            self.uistate.set_active(self.obj.handle, "Media")
        else:
            if self.data_has_changed():
                with DbTxn(
                    _("Edit Media Object (%s)") % self.obj.get_description(), self.db
                ) as trans:
                    if not self.obj.get_gramps_id():
                        self.obj.set_gramps_id(self.db.find_next_media_gramps_id())
                    self.db.commit_media(self.obj, trans)

        self._do_close()
        if self.callback:
            self.callback(self.obj)

    def data_has_changed(self):
        """
        A date comparison can fail incorrectly because we have made the
        decision to store entered text in the date. However, there is no
        entered date when importing from a XML file, so we can get an
        incorrect fail.
        """

        if self.db.readonly:
            return False
        elif self.obj.handle:
            orig = self.get_from_handle(self.obj.handle)
            if orig:
                cmp_obj = orig
            else:
                cmp_obj = self.empty_object()
            return cmp_obj.serialize(True)[1:] != self.obj.serialize(True)[1:]
        else:
            cmp_obj = self.empty_object()
            return cmp_obj.serialize(True)[1:] != self.obj.serialize()[1:]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008  Donald N. Allingham
# Copyright (C) 2009  Douglas S. Blank
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide a simplified table creation interface
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import pickle

# -------------------------------------------------------------------------
#
# GNOME modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.simple import SimpleTable
from gramps.gen.errors import WindowActiveError
from ...utils import model_to_text, text_to_clipboard
from ...widgets.multitreeview import MultiTreeView
from ...ddtargets import DdTargets
from ..quick import run_quick_report_by_name
from ...editors import (
    EditPerson,
    EditEvent,
    EditFamily,
    EditCitation,
    EditSource,
    EditPlace,
    EditRepository,
    EditNote,
    EditMedia,
)


# -------------------------------------------------------------------------
#
# QuickTable class
#
# -------------------------------------------------------------------------
class QuickTable(SimpleTable):
    """
    Provide a simplified table creation interface.
    """

    def set_callback(self, which, callback):
        """
        Override (or add) a function for click/double-click
        """
        if which == "leftclick":
            self._callback_leftclick = callback
        elif which == "leftdouble":
            self._callback_leftdouble = callback

    def button_press_event(self, treeview, event):
        wid = treeview.get_toplevel()
        try:
            winmgr = self.simpledoc.doc.uistate.gwm
            self.track = winmgr.get_item_from_window(wid).track
        except:
            self.track = []
        index = None
        if type(event) == bool:  # enter
            selection = treeview.get_selection()
            store, paths = selection.get_selected_rows()
            tpath = paths[0] if len(paths) > 0 else None
            node = store.get_iter(tpath) if tpath else None
            if node:
                treeview.grab_focus()
                index = store.get_value(node, 0)
        elif event.button == 3:
            x = int(event.x)
            y = int(event.y)
            path_info = treeview.get_path_at_pos(x, y)
            if path_info is not None:
                path, col, cellx, celly = path_info
                selection = treeview.get_selection()
                store, paths = selection.get_selected_rows()
                tpath = paths[0] if len(paths) > 0 else None
                node = store.get_iter(tpath) if tpath else None
                if path:
                    treeview.grab_focus()
                    treeview.set_cursor(path, col, 0)
                if store and node:
                    index = store.get_value(node, 0)  # index Below,
        if index is not None:
            if self._link[index]:
                objclass, handle = self._link[index]
            else:
                return False
            if self.simpledoc.doc.uistate.get_export_mode() and objclass != "Filter":
                return False  # avoid edition during export
            self.popup = Gtk.Menu()
            popup = self.popup
            menu_item = Gtk.MenuItem(label=_("Copy all"))
            menu_item.connect(
                "activate",
                lambda widget: text_to_clipboard(model_to_text(treeview.get_model())),
            )
            popup.append(menu_item)
            menu_item.show()
            # Now add more items to popup menu, if available
            # See details (edit, etc):
            menu_item = Gtk.MenuItem(
                label=_("See %s details", "the object")
                % glocale.trans_objclass(objclass)
            )
            menu_item.connect(
                "activate", lambda widget: self.on_table_doubleclick(treeview)
            )
            popup.append(menu_item)
            menu_item.show()
            # Add other items to menu:
            if objclass == "Person":
                menu_item = Gtk.MenuItem(
                    label=_("Make %s active", "the object")
                    % glocale.trans_objclass("Person")
                )
                menu_item.connect(
                    "activate", lambda widget: self.on_table_click(treeview)
                )
                popup.append(menu_item)
                menu_item.show()
            if self.simpledoc.doc.dbstate.db != self.simpledoc.doc.dbstate.db.basedb:
                if objclass == "Filter" and handle[0] in [
                    "Person",
                    "Family",
                    "Place",
                    "Event",
                    "Repository",
                    "Note",
                    "Media",
                    "Citation",
                    "Source",
                ]:
                    menu_item = Gtk.MenuItem(label=_("See data not in Filter"))
                    menu_item.connect(
                        "activate", lambda widget: self.show_not_in_filter(handle[0])
                    )
                    popup.append(menu_item)
                    menu_item.show()
            # Show the popup menu:
            popup.popup_at_pointer(event)
            return True
        return False

    def show_not_in_filter(self, obj_class):
        run_quick_report_by_name(
            self.simpledoc.doc.dbstate,
            self.simpledoc.doc.uistate,
            "filterbyname",
            "Inverse %s" % obj_class,
            track=self.track,
        )

    def on_table_doubleclick(self, obj):
        """
        Handle events on tables. obj is a treeview
        """
        selection = obj.get_selection()
        store, paths = selection.get_selected_rows()
        tpath = paths[0] if len(paths) > 0 else None
        node = store.get_iter(tpath) if tpath else None
        if not node:
            return
        index = store.get_value(node, 0)  # index
        if self._callback_leftdouble:
            self._callback_leftdouble(store.get_value(node, 1))
            return True
        elif self._link[index]:
            objclass, handle = self._link[index]
            if isinstance(handle, list):
                handle = handle[0]
            if objclass == "Person":
                person = self.access.dbase.get_person_from_handle(handle)
                if person:
                    try:
                        EditPerson(
                            self.simpledoc.doc.dbstate,
                            self.simpledoc.doc.uistate,
                            [],
                            person,
                        )
                        return True  # handled event
                    except WindowActiveError:
                        pass
            elif objclass == "Event":
                event = self.access.dbase.get_event_from_handle(handle)
                if event:
                    try:
                        EditEvent(
                            self.simpledoc.doc.dbstate,
                            self.simpledoc.doc.uistate,
                            [],
                            event,
                        )
                        return True  # handled event
                    except WindowActiveError:
                        pass
            elif objclass == "Family":
                ref = self.access.dbase.get_family_from_handle(handle)
                if ref:
                    try:
                        EditFamily(
                            self.simpledoc.doc.dbstate,
                            self.simpledoc.doc.uistate,
                            [],
                            ref,
                        )
                        return True  # handled event
                    except WindowActiveError:
                        pass
            elif objclass == "Citation":
                ref = self.access.dbase.get_citation_from_handle(handle)
                if ref:
                    try:
                        EditCitation(
                            self.simpledoc.doc.dbstate,
                            self.simpledoc.doc.uistate,
                            [],
                            ref,
                        )
                        return True  # handled event
                    except WindowActiveError:
                        pass
            elif objclass == "Source":
                ref = self.access.dbase.get_source_from_handle(handle)
                if ref:
                    try:
                        EditSource(
                            self.simpledoc.doc.dbstate,
                            self.simpledoc.doc.uistate,
                            [],
                            ref,
                        )
                        return True  # handled event
                    except WindowActiveError:
                        pass
            elif objclass == "Place":
                ref = self.access.dbase.get_place_from_handle(handle)
                if ref:
                    try:
                        EditPlace(
                            self.simpledoc.doc.dbstate,
                            self.simpledoc.doc.uistate,
                            [],
                            ref,
                        )
                        return True  # handled event
                    except WindowActiveError:
                        pass
            elif objclass == "Repository":
                ref = self.access.dbase.get_repository_from_handle(handle)
                if ref:
                    try:
                        EditRepository(
                            self.simpledoc.doc.dbstate,
                            self.simpledoc.doc.uistate,
                            [],
                            ref,
                        )
                        return True  # handled event
                    except WindowActiveError:
                        pass
            elif objclass == "Note":
                ref = self.access.dbase.get_note_from_handle(handle)
                if ref:
                    try:
                        EditNote(
                            self.simpledoc.doc.dbstate,
                            self.simpledoc.doc.uistate,
                            [],
                            ref,
                        )
                        return True  # handled event
                    except WindowActiveError:
                        pass
            elif objclass == "Media":
                ref = self.access.dbase.get_media_from_handle(handle)
                if ref:
                    try:
                        EditMedia(
                            self.simpledoc.doc.dbstate,
                            self.simpledoc.doc.uistate,
                            [],
                            ref,
                        )
                        return True  # handled event
                    except WindowActiveError:
                        pass
            elif objclass == "PersonList":
                run_quick_report_by_name(
                    self.simpledoc.doc.dbstate,
                    self.simpledoc.doc.uistate,
                    "filterbyname",
                    "list of people",
                    handle=handle,
                    track=self.track,
                )
            elif objclass == "Filter":
                if isinstance(handle, list):
                    handle = handle[0]
                run_quick_report_by_name(
                    self.simpledoc.doc.dbstate,
                    self.simpledoc.doc.uistate,
                    "filterbyname",
                    handle,
                    track=self.track,
                )
        return False  # didn't handle event

    def on_table_click(self, obj):
        """
        Handle events on tables. obj is a treeview
        """
        selection = obj.get_selection()
        store, paths = selection.get_selected_rows()
        tpath = paths[0] if len(paths) > 0 else None
        node = store.get_iter(tpath)
        if not node:
            return
        index = store.get_value(node, 0)  # index
        if self._callback_leftclick:
            self._callback_leftclick(store.get_value(node, 1))
            return True
        elif self._link[index]:
            objclass, handle = self._link[index]
            if isinstance(handle, list):
                handle = handle[0]
            if objclass == "Person":
                from gi.repository import GLib

                # If you emmit the signal here and it causes this table to be deleted,
                # then you'll crash Python:
                # self.simpledoc.doc.uistate.set_active(handle, 'Person')
                # So, let's return from this, then change the active person:
                return GLib.timeout_add(
                    100, self.simpledoc.doc.uistate.set_active, handle, "Person"
                )
                return True
        return False  # didn't handle event

    def object_drag_data_get(self, widget, context, sel_data, info, time):
        tree_selection = widget.get_selection()
        model, paths = tree_selection.get_selected_rows()
        retval = []
        for path in paths:
            node = model.get_iter(path)
            index = model.get_value(node, 0)
            if index is not None and self._link[index]:
                retval.append(self._link[index])
        sel_data.set(DdTargets.HANDLE_LIST.atom_drag_type, 8, pickle.dumps(retval))
        return True

    def toggle(self, obj, path, col):
        """
        obj - column widget
        path - row
        col - column
        """
        self.treeview.get_model()[path][col] = not self.treeview.get_model()[path][col]

    def write(self, document):
        self.simpledoc = document
        buffer = self.simpledoc.doc.buffer
        text_view = self.simpledoc.doc.text_view
        text_view.set_sensitive(False)
        model_index = 1  # start after index
        if self._sort_col:
            sort_index = self._columns.index(self._sort_col)
        else:
            sort_index = 0
        treeview = MultiTreeView()
        treeview.enable_model_drag_source(
            Gdk.ModifierType.BUTTON1_MASK,
            [DdTargets.HANDLE_LIST.target()],
            Gdk.DragAction.COPY,
        )
        # treeview.enable_model_drag_dest(DdTargets.all_targets(),
        #                                Gdk.DragAction.DEFAULT)
        treeview.connect("drag_data_get", self.object_drag_data_get)
        treeview.set_grid_lines(Gtk.TreeViewGridLines.BOTH)
        # treeview.connect('row-activated', on_table_doubleclick, self)
        # treeview.connect('cursor-changed', on_table_click, self)
        treeview.connect("button-press-event", self.button_press_event)
        treeview.connect("select-cursor-row", self.button_press_event)
        renderer = Gtk.CellRendererText()
        types = [int]  # index
        cnt = 0
        sort_data = []
        sort_data_types = []
        for col in self._columns:
            if self.get_cell_type(cnt) == "text":
                types.append(str)
                if self.get_cell_markup(cnt):
                    column = Gtk.TreeViewColumn(col, renderer, markup=model_index)
                else:
                    column = Gtk.TreeViewColumn(col, renderer, text=model_index)
            elif self.get_cell_type(cnt) == "checkbox":
                types.append(bool)
                toggle_renderer = Gtk.CellRendererToggle()
                toggle_renderer.set_property("activatable", True)
                toggle_renderer.connect("toggled", self.toggle, model_index)
                column = Gtk.TreeViewColumn(col, toggle_renderer)
                column.add_attribute(toggle_renderer, "active", model_index)
            column.set_resizable(True)
            if self._sort_vals[cnt] != []:
                sort_data.append(self._sort_vals[cnt])
                column.set_sort_column_id(len(self._columns) + len(sort_data))
                sort_data_types.append(int)
            else:
                column.set_sort_column_id(model_index)
            treeview.append_column(column)
            self.model_index_of_column[col] = model_index
            # if model_index == sort_index:
            # FIXME: what to set here?
            model_index += 1
            cnt += 1
        if self.title:
            self.simpledoc.paragraph(self.title)
        # Make a GUI to put the tree view in
        types += sort_data_types
        model = Gtk.ListStore(*types)
        treeview.set_model(model)
        treeview.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)
        iter = buffer.get_end_iter()
        anchor = buffer.create_child_anchor(iter)
        text_view.add_child_at_anchor(treeview, anchor)
        self.treeview = treeview
        count = 0
        for data in self._rows:
            col = 0
            rowdata = []
            for cell in data:
                rowdata.append(self.get_cell_markup(col, count, cell))
                col += 1
            try:
                model.append(
                    row=([count] + list(rowdata) + [col[count] for col in sort_data])
                )
            except KeyError as msg:
                print(msg)
                if sort_data:
                    print(
                        "Quicktable: error in row %d: data: %s, sort data: %d"
                        % (count, rowdata, len(sort_data[0]))
                    )
                else:
                    print("Quicktable: error in row %d: data: %s" % (count, rowdata))
            count += 1
        text_view.show_all()
        self.simpledoc.paragraph("")
        self.simpledoc.paragraph("")
        while Gtk.events_pending():
            Gtk.main_iteration()
        text_view.set_sensitive(True)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010  Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
The User class provides basic interaction with the user.
"""

import sys
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager


class UserBase(metaclass=ABCMeta):
    """
    This class provides a means to interact with the user in an abstract way.
    This class should be overridden by each respective user interface to
    provide the appropriate interaction (eg. dialogs for GTK, prompts for CLI).
    """

    def __init__(self, callback=None, error=None, uistate=None, dbstate=None):
        self.callback_function = callback
        self.error_function = error
        self._fileout = sys.stderr  # redirected to mocks by unit tests
        self.uistate = uistate
        self.dbstate = dbstate

    @abstractmethod
    def begin_progress(self, title, message, steps):
        """
        Start showing a progress indicator to the user.

        Don't use this method directly, use progress instead.

        :param title: the title of the progress meter
        :type title: str
        :param message: the message associated with the progress meter
        :type message: str
        :param steps: the total number of steps for the progress meter.
            a value of 0 indicates that the ending is unknown and the
            meter should just show activity.
        :type steps: int
        :returns: none
        """

    @abstractmethod
    def step_progress(self):
        """
        Advance the progress meter.

        Don't use this method directly, use progress instead.
        """

    def callback(self, percentage, text=None):
        """
        Display the precentage.
        """
        if self.callback_function:
            if text:
                self.callback_function(percentage, text)
            else:
                self.callback_function(percentage)
        else:
            self._default_callback(percentage, text)

    def _default_callback(self, percentage, text):
        if text is None:
            self._fileout.write("\r%02d%%" % percentage)
        else:
            self._fileout.write("\r%02d%% %s" % (percentage, text))

    @abstractmethod
    def end_progress(self):
        """
        Stop showing the progress indicator to the user.

        Don't use this method directly, use progress instead.
        """

    # Context-manager wrapper of the begin/step/end_progress above
    @contextmanager
    def progress(self, *args, **kwargs):
        """
        Preferred form of progress reporting.

        Parameters: same as for begin_progress.

        Usage example (see gramps/cli/test/user_test.py)::

            with self.user.progress("Foo", "Bar", 0) as step:
                for i in range(10):
                    step()

        Ensures end_progress will be called even if an exception was thrown.
        """
        self.begin_progress(*args, **kwargs)
        try:
            yield self.step_progress
        except:
            raise
        finally:
            self.end_progress()

    @abstractmethod
    def prompt(
        self,
        title,
        message,
        accept_label,
        reject_label,
        parent=None,
        default_label=None,
    ):
        """
        Prompt the user with a message to select an alternative.

        :param title: the title of the question, e.g.: "Undo history warning"
        :type title: str
        :param message: the message, e.g.: "Proceeding with the tool will
            erase the undo history. If you think you may want to revert
            running this tool, please stop here and make a backup of the DB."
        :type question: str
        :param accept_label: what to call the positive choice, e.g.: "Proceed"
        :type accept_label: str
        :param reject_label: what to call the negative choice, e.g.: "Stop"
        :type reject_label: str
        :param default_label: the label of the default
        :type default_label: str or None
        :returns: the user's answer to the question
        :rtype: bool
        """

    @abstractmethod
    def warn(self, title, warning=""):
        """
        Warn the user.

        :param title: the title of the warning
        :type title: str
        :param warning: the warning
        :type warning: str
        :returns: none
        """

    @abstractmethod
    def notify_error(self, title, error=""):
        """
        Notify the user of an error.

        :param title: the title of the error
        :type title: str
        :param error: the error message
        :type error: str
        :returns: none
        """

    @abstractmethod
    def notify_db_error(self, error):
        """
        Notify the user of a DB error.

        :param error: the error message
        :type error: str
        :returns: none
        """

    @abstractmethod
    def notify_db_repair(self, error):
        """
        Notify the user their DB might need repair.

        :param error: the error message
        :type error: str
        :returns: none
        """

    @abstractmethod
    def info(self, msg1, infotext, parent=None, monospaced=False):
        """
        Displays information to the user
        """


class User(UserBase):
    """
    An implementation of the :class:`.gen.user.UserBase` class which supresses
    output and accepts prompts.  This is useful for unit tests.
    """

    def __init__(self, callback=None, error=None, uistate=None, dbstate=None):
        UserBase.__init__(self, callback=self.__cb)

    def __cb(self, percent, text=None):
        return

    def begin_progress(self, title, message, steps):
        pass

    def step_progress(self):
        pass

    def end_progress(self):
        pass

    def prompt(
        self,
        title,
        message,
        accept_label,
        reject_label,
        parent=None,
        default_label=None,
    ):
        return True

    def warn(self, title, warning=""):
        pass

    def notify_error(self, title, error=""):
        pass

    def notify_db_error(self, error):
        pass

    def notify_db_repair(self, error):
        pass

    def info(self, msg1, infotext, parent=None, monospaced=False):
        pass



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


import logging

from ._errorview import ErrorView


class GtkHandler(logging.Handler):
    """
    A handler class which pops up a Gtk Window when a log message occurs.
    """

    def __init__(self, rotate_handler=None):
        """
        Initialize the handler with a optional rotate_logger instance.
        """
        logging.Handler.__init__(self)
        self._rotate_handler = rotate_handler

    def emit(self, record):
        """
        Add the record to the rotating buffer.
        """
        self._record = record
        ErrorView(error_detail=self, rotate_handler=self._rotate_handler)

    def get_formatted_log(self):
        return self.format(self._record)

    def get_record(self):
        return self._record
